!*******************************************************************************
!                Time-stamp: <2005-01-18 17:58:51 sander>
!*******************************************************************************

! Authors:
! Rolf Sander,    MPICH, 2003-2004: original code
! Astrid Kerkweg, MPICH, 2003:      inclusion of halogen/mbl chemistry

!*****************************************************************************

! This program is free software; you can redistribute it and/or
! modify it under the terms of the GNU General Public License
! as published by the Free Software Foundation; either version 2
! of the License, or (at your option) any later version.

! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.

! You should have received a copy of the GNU General Public License
! along with this program; if not, get it from:
! http://www.gnu.org/copyleft/gpl.html

!*****************************************************************************

MODULE messy_mecca_box

  USE messy_mecca_kpp_g_mem     ! dp, kpp_*, ...
  USE messy_main_constants_mem, ONLY: R_gas, N_A
  USE messy_mecca,              ONLY: lmbl
  USE messy_mecca_mbl,          ONLY: hrate_diff
  USE mo_netcdf,                ONLY: open_nc_file, write_nc_file, close_nc_file

  IMPLICIT NONE

  INTEGER :: ncid_messy, ncid_mbl, ncid_tracer, ncid_hrate
  REAL(dp) :: time_step_len
  REAL(dp), PUBLIC :: model_time
  REAL(dp) :: zrad_A(AS:CS)     ! radii of accumulation and coarse mode

  ! MXXX = molar mass of element XXX [kg/mol];
  REAL(dp), PARAMETER :: MH  =   1.01E-3 
  REAL(dp), PARAMETER :: MC  =  12.01E-3 
  REAL(dp), PARAMETER :: MN  =  14.01E-3 
  REAL(dp), PARAMETER :: MNa =  22.99E-3 
  REAL(dp), PARAMETER :: MO  =  16.00E-3
  REAL(dp), PARAMETER :: MS  =  32.07E-3 
  REAL(dp), PARAMETER :: MF  =  19.00E-3 
  REAL(dp), PARAMETER :: MCl =  35.45E-3 
  REAL(dp), PARAMETER :: MBr =  79.90E-3 
  REAL(dp), PARAMETER :: MI  = 126.90E-3
  REAL(dp), PARAMETER :: MH2O   = MH*2.+MO
  REAL(dp), PARAMETER :: MH2SO4 = MH*2.+MS+MO*4.

  REAL(dp), PARAMETER :: RHOH2O  = 999.9668 ! density of water [kg/m3]
  REAL(dp), PARAMETER :: OneDay  = 86400.
  REAL(dp), PARAMETER :: InitSpr = 80.               ! first day of spring [d]
  REAL(dp), PARAMETER :: Cancer  = 23.45 * (pi/180.) ! angle of earth
  REAL(dp), PARAMETER :: latitu  = 45. * (pi/180.)   ! latitude

  ! sea water composition, relative to total salt
  REAL(dp), PARAMETER :: Brm_rel   = 1.5E-3
  REAL(dp), PARAMETER :: HCO3m_rel = 4.2E-3
  REAL(dp), PARAMETER :: Im_rel    = 7.4E-8/0.545
  REAL(dp), PARAMETER :: IO3m_rel  = 2.64E-7/0.545
  REAL(dp), PARAMETER :: Clm_rel   = 1.-Brm_rel-HCO3m_rel-Im_rel-IO3m_rel

  REAL(dp) :: relhum, zmbl
  REAL(dp) :: csalt(2), exchng(2)
  CHARACTER(LEN=20) :: species(nspec)

  REAL(dp) :: &
    hrate_NO2_as,     hrate_HNO3_as,    hrate_NH3_as,     hrate_SO2_as,     &
    hrate_H2SO4_as,   hrate_HCOOH_as,   hrate_O3_as,      hrate_HCHO_as,    &
    hrate_H2O2_as,    hrate_HONO_as,    hrate_HNO4_as,    hrate_NO3_as,     &
    hrate_OH_as,      hrate_HO2_as,     hrate_CH3O2_as,   hrate_CO2_as,     &
    hrate_O2_as,      hrate_CH3OOH_as,  hrate_HCl_as,     hrate_HOCl_as,    &
    hrate_Cl2_as,     hrate_Br2_as,     hrate_BrCl_as,    hrate_HBr_as,     &
    hrate_HOBr_as,    hrate_DMSO_as,    hrate_HOI_as,     hrate_IO_as,      &
    hrate_I2_as,      hrate_ICl_as,     hrate_IBr_as,                       &
    hrate_NO2_cs,     hrate_HNO3_cs,    hrate_NH3_cs,     hrate_SO2_cs,     &
    hrate_H2SO4_cs,   hrate_HCOOH_cs,   hrate_O3_cs,      hrate_HCHO_cs,    &
    hrate_H2O2_cs,    hrate_HONO_cs,    hrate_HNO4_cs,    hrate_NO3_cs,     &
    hrate_OH_cs,      hrate_HO2_cs,     hrate_CH3O2_cs,   hrate_CO2_cs,     &
    hrate_O2_cs,      hrate_CH3OOH_cs,  hrate_HCl_cs,     hrate_HOCl_cs,    &
    hrate_Cl2_cs,     hrate_Br2_cs,     hrate_BrCl_cs,    hrate_HBr_cs,     &
    hrate_HOBr_cs,    hrate_DMSO_cs,    hrate_HOI_cs,     hrate_IO_cs,      &
    hrate_I2_cs,      hrate_ICl_cs,     hrate_IBr_cs

  PRIVATE
  PUBLIC :: mecca_init   ! initialize mbl chemistry
  PUBLIC :: mecca_physc  ! calculate physics and chemistry
  PUBLIC :: mecca_result ! print results
  PUBLIC :: mecca_finish ! close files

CONTAINS

  !*****************************************************************************

  SUBROUTINE mecca_init

    USE messy_mecca, ONLY: mecca_read_nml_ctrl, &
                           initialize_kpp_variables

    IMPLICIT NONE

    CHARACTER(LEN=20)   :: unit(nspec)
    CHARACTER(LEN=20)   :: hrate_unit(62)
    INTEGER, PARAMETER  :: iou = 999   ! I/O unit
    INTEGER             :: status ! error status

    model_time = 80. * OneDay      ! time start
    !tend = model_time + OneDay     ! time end
    tend = model_time + 8.* OneDay ! time end
    time_step_len = 20.*60. ! output interval

    ! initialize main-ctrl
    CALL mecca_read_nml_ctrl(status, iou)
    IF (status /= 0) STOP

    CALL initialize_kpp_variables

    ! initialize physics
    CALL update_physc

    CALL x0MARINE        ! initial gas phase mixing ratios

    IF (lmbl) CALL define_aerosol  ! define radius, lwc, etc.

    ! --------------------------------------------------------------------------

    ! open netcdf files and write headers

    CALL open_nc_file(ncid_messy, 'mecca_messy', &
      (/ 'press', 'tm1  ', 'tte  ', 'dt   ' /), &
      (/ 'Pa ', 'K  ', 'K/s', 's  ' /) )

    CALL open_nc_file(ncid_mbl, 'mecca_mbl', &
      (/ 'lwcas_3d', 'lwccs_3d' /), &
      (/ 'm3/m3', 'm3/m3' /) )

    ! define species' names
    INCLUDE "species.inc"
    ! define species' units
    unit(:) = 'mol/mol'
    CALL open_nc_file(ncid_tracer, 'mecca_tracer', species, unit)

    IF (lmbl) THEN
      ! define units for hrate
      hrate_unit(:) = 'mol/mol/day'
      CALL open_nc_file(ncid_hrate, 'mecca_hrate', (/ &
        'hrate_NO2_as   ', 'hrate_HNO3_as  ', 'hrate_NH3_as   ', &
        'hrate_SO2_as   ', 'hrate_H2SO4_as ', 'hrate_HCOOH_as ', &
        'hrate_O3_as    ', 'hrate_HCHO_as  ', 'hrate_H2O2_as  ', &
        'hrate_HONO_as  ', 'hrate_HNO4_as  ', 'hrate_NO3_as   ', &
        'hrate_OH_as    ', 'hrate_HO2_as   ', 'hrate_CH3O2_as ', &
        'hrate_CO2_as   ', 'hrate_O2_as    ', 'hrate_CH3OOH_as', &
        'hrate_HCl_as   ', 'hrate_HOCl_as  ', 'hrate_Cl2_as   ', &
        'hrate_Br2_as   ', 'hrate_BrCl_as  ', 'hrate_HBr_as   ', &
        'hrate_HOBr_as  ', 'hrate_DMSO_as  ', 'hrate_HOI_as   ', &
        'hrate_IO_as    ', 'hrate_I2_as    ', 'hrate_ICl_as   ', &
        'hrate_IBr_as   ', 'hrate_NO2_cs   ', 'hrate_HNO3_cs  ', &
        'hrate_NH3_cs   ', 'hrate_SO2_cs   ', 'hrate_H2SO4_cs ', &
        'hrate_HCOOH_cs ', 'hrate_O3_cs    ', 'hrate_HCHO_cs  ', &
        'hrate_H2O2_cs  ', 'hrate_HONO_cs  ', 'hrate_HNO4_cs  ', &
        'hrate_NO3_cs   ', 'hrate_OH_cs    ', 'hrate_HO2_cs   ', &
        'hrate_CH3O2_cs ', 'hrate_CO2_cs   ', 'hrate_O2_cs    ', &
        'hrate_CH3OOH_cs', 'hrate_HCl_cs   ', 'hrate_HOCl_cs  ', &
        'hrate_Cl2_cs   ', 'hrate_Br2_cs   ', 'hrate_BrCl_cs  ', &
        'hrate_HBr_cs   ', 'hrate_HOBr_cs  ', 'hrate_DMSO_cs  ', &
        'hrate_HOI_cs   ', 'hrate_IO_cs    ', 'hrate_I2_cs    ', &
        'hrate_ICl_cs   ', 'hrate_IBr_cs   '  /),                &
        hrate_unit)
    END IF

    ! --------------------------------------------------------------------------

  END SUBROUTINE mecca_init

  !*****************************************************************************

  SUBROUTINE define_aerosol

    ! radius [m]
    zrad_A(as)=0.0882E-6 ! value from MOCCA
    zrad_A(cs)=1.67E-6   ! value from MOCCA
    ! liquid water content [m3/m3]
    lwc(as) = 1.08E-12 ! value from MOCCA
    lwc(cs) = 3.04E-11 ! value from MOCCA
    ! aerosol on/off
    xaer(as) = 1.
    xaer(cs) = 1.
    ! cvfac: conversion factor dm^3(aq)/mol => cm^3(air)/molecule
    cvfac(as) = 1.E3 / ( N_A * lwc(as) )
    cvfac(cs) = 1.E3 / ( N_A * lwc(cs) )

    ! concentrations
    ! total salt concentration [mcl/cm3(air)]
    csalt(as) = 4.04E-9 * N_A / 1E6 ! value from MOCCA
    csalt(cs) = 1.61E-7 * N_A / 1E6 ! value from MOCCA
    ! sulfate aerosol (all values from MOCCA)
    IF (kpp_H2O_as   /= 0) c(kpp_H2O_as)   = rhoh2o * lwc(as) / MH2O * N_A/1.E6
    IF (kpp_HSO4m_as /= 0) c(kpp_HSO4m_as) = csalt(as)
    IF (kpp_NH4p_as  /= 0) c(kpp_NH4p_as)  = csalt(as)
    exchng(as) = 1. / (7.*OneDay)
    ! sea-salt aerosol (all values from MOCCA)
    IF (kpp_H2O_cs   /= 0) c(kpp_H2O_cs)   = rhoh2o * lwc(cs) / MH2O * N_A/1.E6
    IF (KPP_Nap_cs   /= 0) c(KPP_Nap_cs)   = csalt(cs)
    IF (kpp_HCO3m_cs /= 0) c(kpp_HCO3m_cs) = HCO3m_rel * csalt(cs)
    IF (kpp_Clm_cs   /= 0) c(kpp_Clm_cs)   = Clm_rel   * csalt(cs)
    IF (kpp_Brm_cs   /= 0) c(kpp_Brm_cs)   = Brm_rel   * csalt(cs)
    IF (kpp_Im_cs    /= 0) c(kpp_Im_cs)    = Im_rel    * csalt(cs)
    IF (kpp_IO3m_cs  /= 0) c(kpp_IO3m_cs)  = IO3m_rel  * csalt(cs)
    IF (kpp_Hp_cs    /= 0) c(kpp_Hp_cs)    = 1.e-5 * (N_A/1.e3) * lwc(cs)
    exchng(cs) = 1. / (2.*OneDay)

  END SUBROUTINE define_aerosol

  !*****************************************************************************

  SUBROUTINE x0MARINE

    ! initialize some mixing ratios
    IF (kpp_O2       /= 0) c(kpp_O2)      = 210.E-03 * cair
    IF (kpp_N2       /= 0) c(kpp_N2)      = 780.E-03 * cair
    IF (kpp_H2O      /= 0) c(kpp_H2O)     =  0.01851 * cair
    IF (kpp_O3       /= 0) c(kpp_O3)      =  20.E-09 * cair
    IF (kpp_H2O2     /= 0) c(kpp_H2O2)    = 600.E-12 * cair
    IF (kpp_NH3      /= 0) c(kpp_NH3)     = 100.E-12 * cair
    IF (kpp_NO2      /= 0) c(kpp_NO2)     =  20.E-12 * cair
    IF (kpp_HNO3     /= 0) c(kpp_HNO3)    =  5.0E-12 * cair
    IF (kpp_CH4      /= 0) c(kpp_CH4)     =  1.8E-06 * cair
    IF (kpp_HCHO     /= 0) c(kpp_HCHO)    = 300.E-12 * cair
    IF (kpp_CO       /= 0) c(kpp_CO)      =  70.E-09 * cair
    IF (kpp_CO2      /= 0) c(kpp_CO2)     = 350.E-06 * cair
    IF (kpp_HCl      /= 0) c(kpp_HCl)     =  40.E-12 * cair
    IF (kpp_DMS      /= 0) c(kpp_DMS)     =  60.E-12 * cair
    IF (kpp_SO2      /= 0) c(kpp_SO2)     =  90.E-12 * cair
    IF (kpp_CH3I     /= 0) c(kpp_CH3I)    =   2.E-12 * cair
    IF (kpp_C3H7I    /= 0) c(kpp_C3H7I)   =  1.0E-12 * cair
    IF (kpp_CH3OH    /= 0) c(kpp_CH3OH)   =  1.0E-12 * cair

  END SUBROUTINE x0MARINE

  !*****************************************************************************
  
  SUBROUTINE update_physc

    IMPLICIT NONE

    REAL(dp) :: psat

    temp    = 293.
    press   = 101325.
    relhum  = 0.81
    ! adjust p,T,rh here if they change during model run
    ! temp   = myfunction(model_time)
    ! press  = myfunction(model_time)
    ! relhum = myfunction(model_time)

    zmbl    = 1000.           ! height of boundary layer [m]
    psat    = 6.982616467E+5 & ! saturation partial pressure of gaseous H2O [Pa]
              - 1.888612677E+4*temp    + 2.132971127E+2*temp**2 &
              - 1.288405237E+0*temp**3 + 4.393186046E-3*temp**4 &
              - 8.023554873E-6*temp**5 + 6.136820929E-9*temp**6
    cair    = (N_A/1.E6) * press / (R_gas*temp) ! cair = c(air) in [mcl/cc]
    c(kpp_H2O) = cair * relhum * psat / press

  END SUBROUTINE update_physc

  !*****************************************************************************

  SUBROUTINE mecca_physc

    USE messy_mecca,          ONLY: kpp_integrate
    USE messy_mecca_mbl,      ONLY: mecca_mbl_trans_coeff, &
                                    mecca_mbl_henry, &
                                    mecca_mbl_calc_k_ex

    CALL update_physc
    CALL emis_drydep                     ! emission and dry deposition
    IF (lmbl) THEN
      CALL aerosol_exchng                ! exchange with fresh aerosol
      CALL mecca_mbl_trans_coeff(zrad_A) ! calc. transfer coefficient
      CALL mecca_mbl_henry               ! calc. Henry coefficients
      CALL mecca_mbl_calc_k_ex           ! calc. equilibrium coefficients
      CALL mecca_mbl_calc_hrate          ! diagnostic output of Henry rates
    ENDIF
    CALL update_jval

    CALL check_range('before kpp:',c(:))
    !qqq c(:) = MERGE(c(:),0._dp,c(:) > 0.) ! set negative values to zero
    CALL kpp_integrate(time_step_len)  ! main kpp call
    CALL check_range('after kpp: ',c(:))
    model_time = model_time + time_step_len

  CONTAINS

    !-------------------------------------------------------------------------

    SUBROUTINE check_range(infostring,conc)

      ! print a warning if a concentration is not in the correct range

      CHARACTER(LEN=*), INTENT(IN) :: infostring
      REAL(dp),         INTENT(IN) :: conc(:) ! tracer concentration
      INTEGER :: jt

      INTRINSIC SIZE

      tracer_loop: DO jt=1,SIZE(conc)
        wrong_conc: IF ((conc(jt)<0._dp).OR.(conc(jt)>cair)) THEN
          WRITE(*,'(2A,F10.0,A,1PG12.3E3,2A)') infostring, &
            ', time =', model_time, &
            ', c =', conc(jt), ' mcl/cm3 for ', species(jt)
        ENDIF wrong_conc
      END DO tracer_loop

    END SUBROUTINE check_range

    !-------------------------------------------------------------------------

    SUBROUTINE emis_drydep

      ! Emission and deposition are currently calculated with Euler forward.
      ! This should be changed if numerical problems occur.

      IMPLICIT NONE

      REAL(dp) :: fct

      !-----------------------------------------------------------------------

      ! emissions rates [molec(g)/(cm2*s)] conv. to [molec(g)/cm3]
      fct = time_step_len / (100. * zmbl)
      IF (kpp_O3     /= 0) c(kpp_O3)     = c(kpp_O3)     + 5.E10 * fct
      IF (kpp_NO     /= 0) c(kpp_NO)     = c(kpp_NO)     + 1.5E9 * fct
      IF (kpp_NH3    /= 0) c(kpp_NH3)    = c(kpp_NH3)    + 1.0E9 * fct
      IF (kpp_DMS    /= 0) c(kpp_DMS)    = c(kpp_DMS)    + 2.0E9 * fct
      IF (kpp_CH3I   /= 0) c(kpp_CH3I)   = c(kpp_CH3I)   + 6.0E6 * fct
      IF (kpp_C3H7I  /= 0) c(kpp_C3H7I)  = c(kpp_C3H7I)  + 1.0E7 * fct
      IF (kpp_CH2ClI /= 0) c(kpp_CH2ClI) = c(kpp_CH2ClI) + 2.0E7 * fct
      IF (kpp_CH2I2  /= 0) c(kpp_CH2I2)  = c(kpp_CH2I2)  + 3.0E7 * fct

      !-----------------------------------------------------------------------

      ! deposition velocities [cm/s]
      fct = time_step_len / (zmbl * 100.)
      IF (kpp_O3      /= 0) c(kpp_O3)      = (1.-fct*0.04) * c(kpp_O3)
      IF (kpp_H2O2    /= 0) c(kpp_H2O2)    = (1.-fct*0.5)  * c(kpp_H2O2)
      IF (kpp_NH3     /= 0) c(kpp_NH3)     = (1.-fct*0.1)  * c(kpp_NH3)
      IF (kpp_NO2     /= 0) c(kpp_NO2)     = (1.-fct*0.1)  * c(kpp_NO2)
      IF (kpp_N2O5    /= 0) c(kpp_N2O5)    = (1.-fct*1.0)  * c(kpp_N2O5)
      IF (kpp_HNO3    /= 0) c(kpp_HNO3)    = (1.-fct*2.0)  * c(kpp_HNO3)
      IF (kpp_CH3OOH  /= 0) c(kpp_CH3OOH)  = (1.-fct*0.1)  * c(kpp_CH3OOH)
      IF (kpp_HCHO    /= 0) c(kpp_HCHO)    = (1.-fct*0.5)  * c(kpp_HCHO)
      IF (kpp_HCOOH   /= 0) c(kpp_HCOOH)   = (1.-fct*1.0)  * c(kpp_HCOOH)
      IF (kpp_HCl     /= 0) c(kpp_HCl)     = (1.-fct*2.0)  * c(kpp_HCl)
      IF (kpp_HOCl    /= 0) c(kpp_HOCl)    = (1.-fct*1.0)  * c(kpp_HOCl)
      IF (kpp_ClNO3   /= 0) c(kpp_ClNO3)   = (1.-fct*1.0)  * c(kpp_ClNO3)
      IF (kpp_HBr     /= 0) c(kpp_HBr)     = (1.-fct*2.0)  * c(kpp_HBr)
      IF (kpp_HOBr    /= 0) c(kpp_HOBr)    = (1.-fct*1.0)  * c(kpp_HOBr)
      IF (kpp_BrNO3   /= 0) c(kpp_BrNO3)   = (1.-fct*1.0)  * c(kpp_BrNO3)
      IF (kpp_I2O2    /= 0) c(kpp_I2O2)    = (1.-fct*1.0)  * c(kpp_I2O2)
      IF (kpp_HI      /= 0) c(kpp_HI)      = (1.-fct*1.0)  * c(kpp_HI)
      IF (kpp_HOI     /= 0) c(kpp_HOI)     = (1.-fct*1.0)  * c(kpp_HOI)
      IF (kpp_INO2    /= 0) c(kpp_INO2)    = (1.-fct*1.0)  * c(kpp_INO2)
      IF (kpp_INO3    /= 0) c(kpp_INO3)    = (1.-fct*1.0)  * c(kpp_INO3)
      IF (kpp_SO2     /= 0) c(kpp_SO2)     = (1.-fct*0.5)  * c(kpp_SO2)
      IF (kpp_H2SO4   /= 0) c(kpp_H2SO4)   = (1.-fct*2.0)  * c(kpp_H2SO4)
      IF (kpp_CH3SO3H /= 0) c(kpp_CH3SO3H) = (1.-fct*1.0)  * c(kpp_CH3SO3H)
      IF (kpp_DMSO    /= 0) c(kpp_DMSO)    = (1.-fct*1.0)  * c(kpp_DMSO)

      !-----------------------------------------------------------------------

    END SUBROUTINE emis_drydep

    !-------------------------------------------------------------------------

    SUBROUTINE aerosol_exchng

      ! Aerosol exchange is currently calculated with Euler forward.
      ! This should be changed if numerical problems occur.

      IMPLICIT NONE

      REAL(dp) :: fct_as, fct_cs
      INTEGER :: i

      ! fresh particles
      fct_as = time_step_len * csalt(as) * exchng(as)
      IF (kpp_HSO4m_as /=0) c(kpp_HSO4m_as) = c(kpp_HSO4m_as) + fct_as
      IF (kpp_NH4p_as  /=0) c(kpp_NH4p_as)  = c(kpp_NH4p_as)  + fct_as
      fct_cs = time_step_len * csalt(cs) * exchng(cs)
      IF (kpp_Nap_cs   /=0) c(KPP_Nap_cs)   = c(KPP_Nap_cs)   + fct_cs
      IF (kpp_HCO3m_cs /=0) c(kpp_HCO3m_cs) = c(kpp_HCO3m_cs) + fct_cs*HCO3m_rel
      IF (kpp_Clm_cs   /=0) c(kpp_Clm_cs)   = c(kpp_Clm_cs)   + fct_cs*Clm_rel
      IF (kpp_Brm_cs   /=0) c(kpp_Brm_cs)   = c(kpp_Brm_cs)   + fct_cs*Brm_rel
      IF (kpp_Im_cs    /=0) c(kpp_Im_cs)    = c(kpp_Im_cs)    + fct_cs*Im_rel
      IF (kpp_IO3m_cs  /=0) c(kpp_IO3m_cs)  = c(kpp_IO3m_cs)  + fct_cs*IO3m_rel

      ! particle loss
      fct_as = 1. - time_step_len * exchng(as)
      fct_cs = 1. - time_step_len * exchng(cs)
      DO i = 1,nspec
        ! all species whose names contain '_as' or '_cs' are lost via
        ! particle sedimentation (except for H2O)
        IF ((i==kpp_H2O_as).OR.(i==kpp_H2O_cs)) CYCLE
        IF ((i==kpp_Hp_cs).OR.(i==kpp_OHm_cs)) CYCLE
        IF (INDEX(species(i),'_as') /= 0) c(i) = c(i) * fct_as
        IF (INDEX(species(i),'_cs') /= 0) c(i) = c(i) * fct_cs
      ENDDO

    END SUBROUTINE aerosol_exchng

    !-------------------------------------------------------------------------

    SUBROUTINE update_jval

      IMPLICIT NONE

      INTRINSIC SIN, COS, EXP, LOG

      REAL(dp) :: DayReal, SoDecli, SINPSI, PHOTON, FCT, DN
      REAL(dp), PARAMETER :: DUSK = 0.0721347 ! = 5.E-2/LOG(2.) = PHOTON at dusk

      ! day as a real value (at start of spring DayReal = 0):
      DayReal = (model_time)/OneDay-InitSpr
      ! seasonal cycle:
      SoDecli = Cancer * SIN (2. * PI * DayReal / 365.25)
      ! diurnal cycle of psi, the solar elevation angle
      SINPSI = SIN (Latitu) * SIN (SoDecli) &
      - COS (Latitu) * COS (SoDecli) * COS (2 * PI * DayReal)
      ! PHOTON is approximately the positive part of SINPSI but
      !   - avoid sharp switching on and off
      !   - add some light at dawn before sunrise and at dusk after sunset
      PHOTON = DUSK*LOG(1.+EXP(SINPSI/DUSK))
      ! DN is a day/night switch to set photolyses to about 0 at night
      FCT=50.
      DN = EXP(FCT*SINPSI)/(EXP(-FCT*SINPSI)+EXP(FCT*SINPSI))

      ! J values from PAPER model by Landgraf et al.
      ! J value as a function of PHOTON: J = A*exp(-B/(C+PHOTON))
      ! PHOTON=sin(PSI)=cos(THETA)
      ! THETA=zenith angle, PSI=90-THETA
      ! PHOTON/COS(LATITU) reaches about 1 on noon of first day of spring
      ! temp profile of atmos. is: data/prof.AFGL.midl.sum
      ! surface albedo :  0.00
      ! ozone column (Dobson) :300.00
      ! cloud cover: OFF
      ! paper model was started on: 00-01-18
      J_O1D    = DN * 4.4916E-04 * EXP(-3.5807E+00 /(3.2382E-01+PHOTON)) !J01
      J_O3P    = DN * 4.3917E-04 * EXP(-1.6665E-01 /(5.0375E-02+PHOTON)) !J02
      J_H2O2   = DN * 1.8182E-05 * EXP(-1.2565E+00 /(1.7904E-01+PHOTON)) !J03
      J_NO2    = DN * 1.2516E-02 * EXP(-4.5619E-01 /(6.9151E-02+PHOTON)) !J04
      J_NOO2   = DN * 2.2959E-02 * EXP(-1.0873E-01 /(2.1442E-02+PHOTON)) !J05
      J_NO2O   = DN * 1.9176E-01 * EXP(-1.2557E-01 /(1.9375E-02+PHOTON)) !J06
      J_N2O5   = DN * 1.1375E-04 * EXP(-1.1092E+00 /(1.7004E-01+PHOTON)) !J07
      J_HNO3   = DN * 3.4503E-06 * EXP(-2.1412E+00 /(2.6143E-01+PHOTON)) !J08
      J_CH3OOH = DN * 1.2858E-05 * EXP(-1.1739E+00 /(1.7044E-01+PHOTON)) !J09
      J_CHOH   = DN * 8.6933E-05 * EXP(-1.3293E+00 /(1.6405E-01+PHOTON)) !J10
      J_COH2   = DN * 8.9896E-05 * EXP(-8.4745E-01 /(1.2056E-01+PHOTON)) !J11
      J_HOCl   = DN * 4.9596E-04 * EXP(-8.4303E-01 /(1.3121E-01+PHOTON)) !J12
      J_Cl2O2  = DN * 2.8605E-03 * EXP(-8.8027E-01 /(1.4813E-01+PHOTON)) !J13
      J_ClNO2  = DN * 8.0989E-04 * EXP(-9.7319E-01 /(1.4656E-01+PHOTON)) !J14
      J_ClNO3  = DN * 7.4610E-05 * EXP(-7.2496E-01 /(1.2847E-01+PHOTON)) !J15
      J_Cl2    = DN * 3.9398E-03 * EXP(-6.3121E-01 /(1.0305E-01+PHOTON)) !J16
      J_HOBr   = DN * 3.2114E-03 * EXP(-4.6104E-01 /(8.9870E-02+PHOTON)) !J17
      J_BrNO2  = DN * 7.6992E-03 * EXP(-3.4144E-01 /(6.2243E-02+PHOTON)) !J18
      J_BrNO3  = DN * 1.9670E-03 * EXP(-5.2431E-01 /(9.9522E-02+PHOTON)) !J19
      J_Br2    = DN * 3.6899E-02 * EXP(-2.1097E-01 /(3.3855E-02+PHOTON)) !J20
      J_BrCl   = DN * 1.3289E-02 * EXP(-3.0734E-01 /(5.3134E-02+PHOTON)) !J21
      J_IO     = DN * 3.6977E-01 * EXP(-2.6523E-01 /(3.7541E-02+PHOTON)) !J22
      J_HOI    = DN * 1.2681E-02 * EXP(-3.7515E-01 /(6.1696E-02+PHOTON)) !J23
      J_INO3   = DN * 5.7985E-03 * EXP(-5.2214E-01 /(1.0423E-01+PHOTON)) !J24
      J_CH3I   = DN * 2.6536E-05 * EXP(-1.7937E+00 /(2.4260E-01+PHOTON)) !J25
      J_I2     = DN * 1.6544E-01 * EXP(-1.3475E-01 /(2.1859E-02+PHOTON)) !J26
      J_BrO    = DN * 6.9639E-02 * EXP(-7.4569E-01 /(1.0859E-01+PHOTON)) !J28
      J_ICl    = DN * 2.6074E-02 * EXP(-1.9215E-01 /(3.0177E-02+PHOTON)) !J29
      J_IBr    = DN * 7.4918E-02 * EXP(-1.5931E-01 /(2.3807E-02+PHOTON)) !J30
      J_INO2   = DN * 5.5440E-03 * EXP(-6.6060E-01 /(1.0034E-01+PHOTON)) !J31
      J_C3H7I  = DN * 8.9493E-05 * EXP(-1.8899E+00 /(2.5853E-01+PHOTON)) !J32
      J_CH2ClI = DN * 4.9715E-04 * EXP(-1.4267E+00 /(2.0610E-01+PHOTON)) !J33
      J_CH2I2  = DN * 2.0184E-02 * EXP(-1.0349E+00 /(1.4961E-01+PHOTON)) !J34
      J_OClO   = DN * 1.0933E-01 * EXP(-4.4797E-01 /(7.2470E-02+PHOTON)) !J89
      J_HNO4   = DN * 2.1532E-05 * EXP(-1.9648E+00 /(2.1976E-01+PHOTON)) !J91
      J_HONO   = DN * 2.9165E-03 * EXP(-5.1317E-01 /(7.4940E-02+PHOTON)) !J92
      J_CH3Br  = DN * 7.3959E-10 * EXP(-2.9326E+01 /(1.0132E-01+PHOTON)) !J99

    END SUBROUTINE update_jval
    
    !---------------------------------------------------------------------------
    
    SUBROUTINE mecca_mbl_calc_hrate
      IF (xaer(AS)/=0) THEN

        IF(KPP_NO2    * KPP_NO2_as    /=0 ) &
          hrate_NO2_as   = hrate_diff(AS, KPP_NO2,  KPP_NO2_as)
        IF(KPP_HNO3   * KPP_HNO3_as   /=0 ) &
          hrate_HNO3_as  = hrate_diff(AS, KPP_HNO3, KPP_HNO3_as)
        IF(KPP_NH3    * KPP_NH3_as    /=0) &
          hrate_NH3_as   = hrate_diff(AS, KPP_NH3,  KPP_NH3_as)
        IF(KPP_SO2    * KPP_SO2_as    /=0) &
          hrate_SO2_as   = hrate_diff(AS, KPP_SO2,  KPP_SO2_as)
        IF(KPP_H2SO4  * KPP_H2SO4_as  /=0) &
          hrate_H2SO4_as = hrate_diff(AS, KPP_H2SO4, KPP_H2SO4_as)
        IF(KPP_HCOOH  * KPP_HCOOH_as  /=0) &
          hrate_HCOOH_as = hrate_diff(AS, KPP_HCOOH, KPP_HCOOH_as)
        IF(KPP_O3     * KPP_O3_as     /=0) &
          hrate_O3_as    = hrate_diff(AS, KPP_O3,   KPP_O3_as)
        IF(KPP_HCHO   * KPP_HCHO_as   /=0) &
          hrate_HCHO_as  = hrate_diff(AS, KPP_HCHO, KPP_HCHO_as)
        IF(KPP_H2O2   * KPP_H2O2_as   /=0) &
          hrate_H2O2_as  = hrate_diff(AS, KPP_H2O2, KPP_H2O2_as)
        IF(KPP_HONO   * KPP_HONO_as   /=0) &
          hrate_HONO_as  = hrate_diff(AS, KPP_HONO, KPP_HONO_as)
        IF(KPP_HNO4   * KPP_HNO4_as   /=0) &
          hrate_HNO4_as  = hrate_diff(AS, KPP_HNO4, KPP_HNO4_as)
        IF(KPP_NO3    * KPP_NO3_as    /=0) &
          hrate_NO3_as   = hrate_diff(AS, KPP_NO3,  KPP_NO3_as)
        IF(KPP_OH     * KPP_OH_as     /=0) &
          hrate_OH_as    = hrate_diff(AS, KPP_OH,   KPP_OH_as)
        IF(KPP_HO2    * KPP_HO2_as    /=0) &
          hrate_HO2_as   = hrate_diff(AS, KPP_HO2,  KPP_HO2_as)
        IF(KPP_CH3O2  * KPP_CH3O2_as  /=0) &
          hrate_CH3O2_as = hrate_diff(AS, KPP_CH3O2,  KPP_CH3O2_as)
        IF(KPP_CO2    * KPP_CO2_as    /=0) &
          hrate_CO2_as   = hrate_diff(AS, KPP_CO2,  KPP_CO2_as)
        IF(KPP_O2     * KPP_O2_as     /=0) &
          hrate_O2_as    = hrate_diff(AS, KPP_O2,   KPP_O2_as)
        IF(KPP_CH3OOH * KPP_CH3OOH_as /=0) &
          hrate_CH3OOH_as= hrate_diff(AS, KPP_CH3OOH, KPP_CH3OOH_as)
        IF(KPP_HCl  * KPP_HCl_as    /=0) &
          hrate_HCl_as   = hrate_diff(AS, KPP_HCl,  KPP_HCl_as)
        IF(KPP_HOCl * KPP_HOCl_as   /=0) &
          hrate_HOCl_as  = hrate_diff(AS, KPP_HOCl, KPP_HOCl_as)
        IF(KPP_Cl2  * KPP_Cl2_as    /=0) &
          hrate_Cl2_as   = hrate_diff(AS, KPP_Cl2,  KPP_Cl2_as)
        IF(KPP_Br2  * KPP_Br2_as    /=0) &
          hrate_Br2_as   = hrate_diff(AS, KPP_Br2,  KPP_Br2_as)
        IF(KPP_BrCl * KPP_BrCl_as   /=0) &
          hrate_BrCl_as  = hrate_diff(AS, KPP_BrCl, KPP_BrCl_as)
        IF(KPP_HBr  * KPP_HBr_as    /=0) &
          hrate_HBr_as   = hrate_diff(AS, KPP_HBr,  KPP_HBr_as)
        IF(KPP_HOBr * KPP_HOBr_as   /=0) &
          hrate_HOBr_as  = hrate_diff(AS, KPP_HOBr, KPP_HOBr_as)
        IF(KPP_DMSO * KPP_DMSO_as    /=0) &
          hrate_DMSO_as  = hrate_diff(AS, KPP_DMSO, KPP_DMSO_as)
        IF(KPP_HOI * KPP_HOI_as /=0) &
          hrate_HOI_as   = hrate_diff(AS, KPP_HOI,  KPP_HOI_as)
        IF(KPP_IO  * KPP_IO_as  /=0) &
          hrate_IO_as    = hrate_diff(AS, KPP_IO,   KPP_IO_as)
        IF(KPP_I2  * KPP_I2_as  /=0) &
          hrate_I2_as    = hrate_diff(AS, KPP_I2,   KPP_I2_as)
        IF(KPP_ICl * KPP_ICl_as /=0) &
          hrate_ICl_as   = hrate_diff(AS, KPP_ICl,  KPP_ICl_as)
        IF(KPP_IBr * KPP_IBr_as /=0) &
          hrate_IBr_as   = hrate_diff(AS, KPP_IBr,  KPP_IBr_as)

      ELSE ! no as mode => hrate = 0.

        hrate_NO2_as   = 0.
        hrate_HNO3_as  = 0.
        hrate_NH3_as   = 0.
        hrate_SO2_as   = 0.
        hrate_H2SO4_as = 0.
        hrate_HCOOH_as = 0.
        hrate_O3_as    = 0.
        hrate_HCHO_as  = 0.
        hrate_H2O2_as  = 0.
        hrate_HONO_as  = 0.
        hrate_HNO4_as  = 0.
        hrate_NO3_as   = 0.
        hrate_OH_as    = 0.
        hrate_HO2_as   = 0.
        hrate_CH3O2_as = 0.
        hrate_CO2_as   = 0.
        hrate_O2_as    = 0.
        hrate_CH3OOH_as= 0.
        hrate_HCl_as   = 0.
        hrate_HOCl_as  = 0.
        hrate_Cl2_as   = 0.
        hrate_Br2_as   = 0.
        hrate_BrCl_as  = 0.
        hrate_HBr_as   = 0.
        hrate_HOBr_as  = 0.
        hrate_DMSO_as  = 0.
        hrate_HOI_as   = 0.
        hrate_IO_as    = 0.
        hrate_I2_as    = 0.
        hrate_ICl_as   = 0.
        hrate_IBr_as   = 0.

      ENDIF

      IF (xaer(CS)/=0) THEN

        IF(KPP_NO2    * KPP_NO2_cs    /=0) &
          hrate_NO2_cs   = hrate_diff(CS, KPP_NO2,  KPP_NO2_cs)
        IF(KPP_HNO3   * KPP_HNO3_cs   /=0) &
          hrate_HNO3_cs  = hrate_diff(CS, KPP_HNO3, KPP_HNO3_cs)
        IF(KPP_NH3    * KPP_NH3_cs    /=0) &
          hrate_NH3_cs   = hrate_diff(CS, KPP_NH3,  KPP_NH3_cs)
        IF(KPP_SO2    * KPP_SO2_cs    /=0) &
          hrate_SO2_cs   = hrate_diff(CS, KPP_SO2,  KPP_SO2_cs)
        IF(KPP_H2SO4  * KPP_H2SO4_cs  /=0) &
          hrate_H2SO4_cs = hrate_diff(CS, KPP_H2SO4, KPP_H2SO4_cs)
        IF(KPP_HCOOH  * KPP_HCOOH_cs  /=0) &
          hrate_HCOOH_cs = hrate_diff(CS, KPP_HCOOH, KPP_HCOOH_cs)
        IF(KPP_O3     * KPP_O3_cs     /=0) &
          hrate_O3_cs    = hrate_diff(CS, KPP_O3,   KPP_O3_cs)
        IF(KPP_HCHO   * KPP_HCHO_cs   /=0) &
          hrate_HCHO_cs  = hrate_diff(CS, KPP_HCHO, KPP_HCHO_cs)
        IF(KPP_H2O2   * KPP_H2O2_cs   /=0) &
          hrate_H2O2_cs  = hrate_diff(CS, KPP_H2O2, KPP_H2O2_cs)
        IF(KPP_HONO   * KPP_HONO_cs   /=0) &
          hrate_HONO_cs  = hrate_diff(CS, KPP_HONO, KPP_HONO_cs)
        IF(KPP_HNO4   * KPP_HNO4_cs   /=0) &
          hrate_HNO4_cs  = hrate_diff(CS, KPP_HNO4, KPP_HNO4_cs)
        IF(KPP_NO3    * KPP_NO3_cs    /=0) &
          hrate_NO3_cs   = hrate_diff(CS, KPP_NO3,  KPP_NO3_cs)
        IF(KPP_OH     * KPP_OH_cs     /=0) &
          hrate_OH_cs    = hrate_diff(CS, KPP_OH,   KPP_OH_cs)
        IF(KPP_HO2    * KPP_HO2_cs    /=0) &
          hrate_HO2_cs   = hrate_diff(CS, KPP_HO2,  KPP_HO2_cs)
        IF(KPP_CH3O2  * KPP_CH3O2_cs  /=0) &
          hrate_CH3O2_cs = hrate_diff(CS, KPP_CH3O2,  KPP_CH3O2_cs)
        IF(KPP_CO2    * KPP_CO2_cs    /=0) &
          hrate_CO2_cs   = hrate_diff(CS, KPP_CO2,  KPP_CO2_cs)
        IF(KPP_O2     * KPP_O2_cs     /=0) &
          hrate_O2_cs    = hrate_diff(CS, KPP_O2,   KPP_O2_cs)
        IF(KPP_CH3OOH * KPP_CH3OOH_cs /=0) &
          hrate_CH3OOH_cs= hrate_diff(CS, KPP_CH3OOH, KPP_CH3OOH_cs)
        IF(KPP_HCl  * KPP_HCl_cs    /=0) &
          hrate_HCl_cs   = hrate_diff(CS, KPP_HCl,  KPP_HCl_cs)
        IF(KPP_HOCl * KPP_HOCl_cs   /=0) &
          hrate_HOCl_cs  = hrate_diff(CS, KPP_HOCl, KPP_HOCl_cs)
        IF(KPP_Cl2  * KPP_Cl2_cs    /=0) &
          hrate_Cl2_cs   = hrate_diff(CS, KPP_Cl2,  KPP_Cl2_cs)
        IF(KPP_Br2  * KPP_Br2_cs    /=0) &
          hrate_Br2_cs   = hrate_diff(CS, KPP_Br2,  KPP_Br2_cs)
        IF(KPP_BrCl * KPP_BrCl_cs   /=0) &
          hrate_BrCl_cs  = hrate_diff(CS, KPP_BrCl, KPP_BrCl_cs)
        IF(KPP_HBr  * KPP_HBr_cs    /=0) &
          hrate_HBr_cs   = hrate_diff(CS, KPP_HBr,  KPP_HBr_cs)
        IF(KPP_HOBr * KPP_HOBr_cs   /=0) &
          hrate_HOBr_cs  = hrate_diff(CS, KPP_HOBr, KPP_HOBr_cs)
        IF(KPP_DMSO * KPP_DMSO_cs    /=0) &
          hrate_DMSO_cs  = hrate_diff(CS, KPP_DMSO, KPP_DMSO_cs)
        IF(KPP_HOI * KPP_HOI_cs /=0) &
          hrate_HOI_cs   = hrate_diff(CS, KPP_HOI,  KPP_HOI_cs)
        IF(KPP_IO  * KPP_IO_cs  /=0) &
          hrate_IO_cs    = hrate_diff(CS, KPP_IO,   KPP_IO_cs)
        IF(KPP_I2  * KPP_I2_cs  /=0) &
          hrate_I2_cs    = hrate_diff(CS, KPP_I2,   KPP_I2_cs)
        IF(KPP_ICl * KPP_ICl_cs /=0) &
          hrate_ICl_cs   = hrate_diff(CS, KPP_ICl,  KPP_ICl_cs)
        IF(KPP_IBr * KPP_IBr_cs /=0) &
          hrate_IBr_cs   = hrate_diff(CS, KPP_IBr,  KPP_IBr_cs)

      ELSE ! no cs mode => hrate = 0.

        hrate_NO2_cs   = 0.
        hrate_HNO3_cs  = 0.
        hrate_NH3_cs   = 0.
        hrate_SO2_cs   = 0.
        hrate_H2SO4_cs = 0.
        hrate_HCOOH_cs = 0.
        hrate_O3_cs    = 0.
        hrate_HCHO_cs  = 0.
        hrate_H2O2_cs  = 0.
        hrate_HONO_cs  = 0.
        hrate_HNO4_cs  = 0.
        hrate_NO3_cs   = 0.
        hrate_OH_cs    = 0.
        hrate_HO2_cs   = 0.
        hrate_CH3O2_cs = 0.
        hrate_CO2_cs   = 0.
        hrate_O2_cs    = 0.
        hrate_CH3OOH_cs= 0.
        hrate_HCl_cs   = 0.
        hrate_HOCl_cs  = 0.
        hrate_Cl2_cs   = 0.
        hrate_Br2_cs   = 0.
        hrate_BrCl_cs  = 0.
        hrate_HBr_cs   = 0.
        hrate_HOBr_cs  = 0.
        hrate_DMSO_cs  = 0.
        hrate_HOI_cs   = 0.
        hrate_IO_cs    = 0.
        hrate_I2_cs    = 0.
        hrate_ICl_cs   = 0.
        hrate_IBr_cs   = 0.

      ENDIF

    END SUBROUTINE mecca_mbl_calc_hrate

    !-------------------------------------------------------------------------

  END SUBROUTINE mecca_physc

  !*****************************************************************************

  SUBROUTINE mecca_result

    CALL write_nc_file(ncid_messy, model_time, &
      (/ press, temp, 0.0_dp, time_step_len /) ) ! press, tm1, tte, dt
    CALL write_nc_file(ncid_mbl,    model_time, lwc)
    CALL write_nc_file(ncid_tracer, model_time, c/cair)
    IF (lmbl) CALL write_nc_file(ncid_hrate, model_time, (/                   &
      hrate_NO2_as,     hrate_HNO3_as,    hrate_NH3_as,     hrate_SO2_as,     &
      hrate_H2SO4_as,   hrate_HCOOH_as,   hrate_O3_as,      hrate_HCHO_as,    &
      hrate_H2O2_as,    hrate_HONO_as,    hrate_HNO4_as,    hrate_NO3_as,     &
      hrate_OH_as,      hrate_HO2_as,     hrate_CH3O2_as,   hrate_CO2_as,     &
      hrate_O2_as,      hrate_CH3OOH_as,  hrate_HCl_as,     hrate_HOCl_as,    &
      hrate_Cl2_as,     hrate_Br2_as,     hrate_BrCl_as,    hrate_HBr_as,     &
      hrate_HOBr_as,    hrate_DMSO_as,    hrate_HOI_as,     hrate_IO_as,      &
      hrate_I2_as,      hrate_ICl_as,     hrate_IBr_as,                       &
      hrate_NO2_cs,     hrate_HNO3_cs,    hrate_NH3_cs,     hrate_SO2_cs,     &
      hrate_H2SO4_cs,   hrate_HCOOH_cs,   hrate_O3_cs,      hrate_HCHO_cs,    &
      hrate_H2O2_cs,    hrate_HONO_cs,    hrate_HNO4_cs,    hrate_NO3_cs,     &
      hrate_OH_cs,      hrate_HO2_cs,     hrate_CH3O2_cs,   hrate_CO2_cs,     &
      hrate_O2_cs,      hrate_CH3OOH_cs,  hrate_HCl_cs,     hrate_HOCl_cs,    &
      hrate_Cl2_cs,     hrate_Br2_cs,     hrate_BrCl_cs,    hrate_HBr_cs,     &
      hrate_HOBr_cs,    hrate_DMSO_cs,    hrate_HOI_cs,     hrate_IO_cs,      &
      hrate_I2_cs,      hrate_ICl_cs,     hrate_IBr_cs                        &
    /) )

    PRINT *, 'Time = ', model_time, ' s'

  END SUBROUTINE mecca_result

  !*****************************************************************************

  SUBROUTINE mecca_finish

    CALL close_nc_file(ncid_messy)
    CALL close_nc_file(ncid_mbl)
    CALL close_nc_file(ncid_tracer)
    IF (lmbl) CALL close_nc_file(ncid_hrate)

  END SUBROUTINE mecca_finish

  !*****************************************************************************

END MODULE messy_mecca_box

!*******************************************************************************
