!*******************************************************************************
!                Time-stamp: <2004-12-07 19:21:56 sander>
!*******************************************************************************

! mecca-mbl is a part of submodel MECCA
! Calculates (halogen) aerosol chemistry
! written by Astrid Kerkweg, MPICH, Mainz, 2003
! Rolf Sander, 2003: strict separation between core and e5 files

!*****************************************************************************

! This program is free software; you can redistribute it and/or
! modify it under the terms of the GNU General Public License
! as published by the Free Software Foundation; either version 2
! of the License, or (at your option) any later version.

! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.

! You should have received a copy of the GNU General Public License
! along with this program; if not, get it from:
! http://www.gnu.org/copyleft/gpl.html

!*****************************************************************************

MODULE messy_mecca_mbl

  USE messy_main_constants_mem, ONLY: N_A, R_gas, T0
  USE messy_mecca_kpp_g_mem

  IMPLICIT NONE
  PRIVATE

  CHARACTER(len=*), PARAMETER, PUBLIC :: submodstr='mecca_mbl'

  LOGICAL, PUBLIC :: lhrate  = .FALSE. ! calculate henry rates
  REAL(dp) :: ykmt(AS:CS,0:NSPEC)=0.   ! mass transfer coefficient
  REAL(dp) :: yhenry(0:NSPEC)=0.       ! Inverse Henry constant, dimensionless

  PUBLIC :: mecca_mbl_read_nml_ctrl
  PUBLIC :: hrate_diff
  PUBLIC :: mecca_mbl_trans_coeff
  PUBLIC :: mecca_mbl_henry
  PUBLIC :: mecca_mbl_calc_k_ex
  PUBLIC :: density
  PUBLIC :: layerthickness

CONTAINS

  ! --------------------------------------------------------------------------

  SUBROUTINE mecca_mbl_read_nml_ctrl(status, iou, modstr)

    ! READ MECCA-MBL CTRL-NAMELIST AND INITIALIZE GLOBAL VARIABLES
    !
    ! Author: Astrid Kerkweg, MPICH, August 2004

    USE messy_main_tools, ONLY: read_nml_open, read_nml_check, read_nml_close

    IMPLICIT NONE

    ! I/O
    INTEGER, INTENT(OUT)          :: status ! error status
    INTEGER, INTENT(IN)           :: iou    ! logical I/O unit
    CHARACTER(len=*), INTENT(IN)  :: modstr ! MECCA modstr = NAMELIST NAME

    ! LOCAL
    LOGICAL :: lex   ! file exists?
    INTEGER :: fstat ! file status
    CHARACTER(LEN=*), PARAMETER :: substr = 'mecca_mbl_read_nml_ctrl'

    NAMELIST /CTRL_MBL/  lhrate

    ! INITIALIZE
    status = 1 ! error

    ! INPUT NAMELIST
    CALL read_nml_open(lex, substr, iou, 'CTRL_MBL', modstr)
    IF (.not.lex) RETURN    ! <modstr>.nml does not exist

    READ(iou, NML=CTRL_MBL, IOSTAT=fstat)
    CALL read_nml_check(fstat, substr, iou, 'CTRL_MBL', modstr)
    IF (fstat /= 0) RETURN  ! error while reading namelist

    WRITE(*,*) '******************************************************'
    IF (lhrate) THEN
      WRITE(*,*) 'calculation and storage of henry rates'
    ELSE
      WRITE(*,*) 'no calculation of henry rates'
    END IF
    WRITE(*,*) '******************************************************'

    CALL read_nml_close(substr, iou, modstr)
    status = 0 ! no error

  END SUBROUTINE mecca_mbl_read_nml_ctrl

  ! --------------------------------------------------------------------------
  !***************************************************************************

  SUBROUTINE mecca_mbl_trans_coeff(radius)

    ! 1.) Calculation of the transfer coefficient for the mean mode actual
    ! radius
    ! transfer coefficient after Schwarz, 1986 (see Sander & Crutzen '96)
    ! k_mt=1/(r**2/(3*D_gas)+4*r/(3*v_mean*alpha))
    ! if : D_gas=lambda*v_mean/3. then we can rewrite k_mt as:
    ! k_mt=vmean/(r**2/lambda+4*r/(3*alpha))

    INTEGER  :: jn , zkc
    REAL(dp) :: zx1, zrc_t,  zlambda
    REAL(dp) :: intfac(AS:CS)

    REAL(dp) :: vmean(0:nspec)
    REAL(dp) :: alpha(0:nspec)
    REAL(dp) :: radius(AS:CS)

    ! multiplying factor to make the error of ignoring the need of
    ! integration over the mode minimal with little numerical effort
    ! deduced for box-model study

    intfac(AS) = 0.84_dp
    intfac(CS) = 0.55_dp

    ! zlambda = mean free path
    zlambda = 2.28E-5 * temp / press

    ! calculate mean velocities
    CALL mecca_mbl_vmean(vmean)
    ! calculate accommodation coefficients alpha
    CALL mecca_mbl_alpha(alpha)
    zrc_t=1.E-9_dp                    !radius threshold for kmt calculation
    zx1=0.

    ykmt(:,:)  = 0.
    DO jn=1,nspec
      DO zkc=AS,CS ! aerosol mode (phase)
        IF (radius(zkc).GE.zrc_t.AND.alpha(jn).GT.0.) THEN
          zx1=1./(radius(zkc)*(radius(zkc)/zlambda+4./(3.*alpha(jn))))
        ELSE
          zx1=0.
        ENDIF
        ykmt(zkc,jn)=vmean(jn)*zx1*intfac(zkc)
      ENDDO
    ENDDO
    
    ! 2.) Calculation of mean (integrated) transfer coefficient for each mode
    ! omitted because the numerical effort for the numerical integration is
    ! too high

  CONTAINS

    ! ------------------------------------------------------------------------
    
    SUBROUTINE mecca_mbl_vmean(vmean)

      REAL(dp) :: vmean(0:nspec)

      vmean(:) = -999.999_dp ! dummy value

      vmean(KPP_NO)      = vmean_func(3.E-2_dp)
      vmean(KPP_NO2)     = vmean_func(4.6E-2_dp)
      vmean(KPP_HNO3)    = vmean_func(6.3E-2_dp)
      vmean(KPP_NH3)     = vmean_func(1.7E-2_dp)
      vmean(KPP_SO2)     = vmean_func(6.4E-2_dp)
      vmean(KPP_H2SO4)   = vmean_func(9.8E-2_dp)
      vmean(KPP_O3)      = vmean_func(4.8E-2_dp)
      vmean(KPP_CH4)     = vmean_func(1.6E-2_dp)
      vmean(KPP_C2H6)    = vmean_func(3.E-2_dp)
      vmean(KPP_HCOOH)   = vmean_func(4.6E-2_dp)
      vmean(KPP_HCHO)    = vmean_func(3.E-2_dp)
      vmean(KPP_H2O2)    = vmean_func(3.4E-2_dp)
      vmean(KPP_CH3OOH)  = vmean_func(4.8E-2_dp) ! value for CH3OOH
      vmean(KPP_HONO)    = vmean_func(4.7E-2_dp)
      vmean(KPP_PAN)     = vmean_func(1.21E-1_dp)
      vmean(KPP_N2O5)    = vmean_func(1.08E-1_dp)
      vmean(KPP_HNO4)    = vmean_func(6.3E-2_dp)
      vmean(KPP_NO3)     = vmean_func(6.2E-2_dp)
      vmean(KPP_DMS)     = vmean_func(6.2E-2_dp)
      vmean(KPP_HCl)     = vmean_func(3.6E-2_dp)
      vmean(KPP_HOCl)    = vmean_func(5.2E-2_dp)
      vmean(KPP_ClNO2)   = vmean_func(8.1E-2_dp)
      vmean(KPP_ClNO3)   = vmean_func(9.7E-2_dp)
      vmean(KPP_Cl2)     = vmean_func(7.1E-2_dp)
      vmean(KPP_HBr)     = vmean_func(8.1E-2_dp)
      vmean(KPP_HOBr)    = vmean_func(9.7E-2_dp)
      vmean(KPP_BrNO2)   = vmean_func(1.26E-1_dp)
      vmean(KPP_BrNO3)   = vmean_func(1.42E-1_dp)
      vmean(KPP_Br2)     = vmean_func(1.6E-1_dp)
      vmean(KPP_BrCl)    = vmean_func(1.15E-1_dp)
      vmean(KPP_HI)      = vmean_func(1.28E-1_dp)
      vmean(KPP_HOI)     = vmean_func(1.44E-1_dp)
      vmean(KPP_I2O2)    = vmean_func(2.86E-1_dp)
      vmean(KPP_INO2)    = vmean_func(1.73E-1_dp)
      vmean(KPP_INO3)    = vmean_func(1.89E-1_dp)
      vmean(KPP_I2)      = vmean_func(2.54E-1_dp)
      vmean(KPP_ICl)     = vmean_func(1.62E-1_dp)
      vmean(KPP_IBr)     = vmean_func(2.07E-1_dp)
      vmean(KPP_CH3I)    = vmean_func(1.42E-1_dp)
      vmean(KPP_CH2I2)   = vmean_func(2.68E-1_dp)
      vmean(KPP_CH2ClI)  = vmean_func(1.76E-1_dp)
      vmean(KPP_C3H7I)   = vmean_func(1.7E-1_dp)
      vmean(KPP_DMSO)    = vmean_func(7.8E-2_dp)
      vmean(KPP_CH3SO2)  = vmean_func(7.9E-2_dp)
      vmean(KPP_CH3SO3)  = vmean_func(9.5E-2_dp)
      vmean(KPP_CH3SO3H) = vmean_func(9.6E-2_dp)
      vmean(KPP_CO)      = vmean_func(2.8E-2_dp)
      vmean(KPP_CO2)     = vmean_func(4.4E-2_dp)
      vmean(KPP_OH)      = vmean_func(1.7E-2_dp)
      vmean(KPP_HO2)     = vmean_func(3.3E-2_dp)
      vmean(KPP_CH3O2)   = vmean_func(4.7E-2_dp)
      vmean(KPP_Cl)      = vmean_func(3.5E-2_dp)
      vmean(KPP_ClO)     = vmean_func(5.1E-2_dp)
      vmean(KPP_OClO)    = vmean_func(6.7E-2_dp)
      vmean(KPP_Br)      = vmean_func(8.E-2_dp)
      vmean(KPP_BrO)     = vmean_func(9.6E-2_dp)
      vmean(KPP_I)       = vmean_func(1.27E-1_dp)
      vmean(KPP_IO)      = vmean_func(1.43E-1_dp)
      vmean(KPP_O2)      = vmean_func(3.2E-2_dp)
      vmean(KPP_OIO)     = vmean_func(1.59E-1_dp)
      vmean(KPP_HIO3)    = vmean_func(1.76E-1_dp)

      ! dummies
      ! mz_ak_20030514      vmean(KPP_DGtAig)  = vmean_func(30.1E-3)
      ! mz_ak_20030514      vmean(KPP_DGtAsg)  = vmean_func(30.1E-3)

    END SUBROUTINE mecca_mbl_vmean

    ! ------------------------------------------------------------------------

    REAL(dp) FUNCTION vmean_func(molweight)

      ! mean molecular speed from Maxwell-Boltzmann distribution:
      ! vmean=sqrt(8*R_gas*T/(M*pi))      (M in kg/mol)
      ! vmean in m/s
      ! sqrt(8*R_gas/pi)=4.60138

      REAL(dp) :: molweight

      INTRINSIC SQRT

      vmean_func = SQRT(temp/molweight)*4.60138

    END FUNCTION vmean_func

    ! ------------------------------------------------------------------------

    SUBROUTINE mecca_mbl_alpha(alpha)

      ! calculation of accommodation coefficients needed for kmt calculations
      USE messy_main_constants_mem, ONLY: cal2J

      REAL(dp) :: ztcorr, ztemp
      REAL(dp) :: alpha(0:nspec)

      INTRINSIC EXP

      ztemp= temp
      ! init
      alpha(:)      = 0.1_dp

      ztcorr=1./temp-1./T0
      ! for some species a temperature dependence is calculated/estimated
      ! using:
      ! alpha = 1./(1.+1./(1./(1./alpha(T0)-1.)*EXP((-DeltaH/R_gas)*ztcorr)))

      ! standard value
      alpha(KPP_H2SO4)   = 0.65_dp
      alpha(KPP_HNO4)    = 0.1_dp

      ! experimentally determined values (from MOCCA)
      alpha(KPP_O3)      = 2.0E-03_dp
      alpha(KPP_O2)      = 1.0/(1.0_dp+1.0/(1.0/(1.0/1.0E-2 -1.0_dp)*EXP(2000.*ztcorr)))
      alpha(KPP_OH)      = 1.0E-2_dp
      alpha(KPP_HO2)     = 2.0E-01_dp
      alpha(KPP_H2O2)    = 1./(EXP(-26.E3/(R_gas*ztemp)+107.8456/R_gas) + 1.0_dp)
      alpha(KPP_NO2)     = 1.5E-03_dp
      alpha(KPP_NO3)     = 4.0E-02_dp
      alpha(KPP_N2O5)    = 1.0E-01_dp
      alpha(KPP_HONO)    = 4.0E-02_dp
      alpha(KPP_HNO3)    = 5.0E-01_dp
      alpha(KPP_NH3)     = 6.0E-02_dp
      alpha(KPP_CH3O2)   = 1.0_dp/(1.0_dp+1.0_dp/(1.0_dp/(1.0_dp/1.0E-2 -1.0_dp)* &
                           EXP(2000.*ztcorr)))
      alpha(KPP_CH3OOH)  = 1.0_dp/(EXP(-6.5E3*cal2J/(R_gas*ztemp)+32.5*cal2J/R_gas) + 1.0_dp)
      alpha(KPP_HCHO)    = 4.0E-02_dp
      alpha(KPP_HCOOH)   = 1.0_dp/(EXP(-7.9E3*cal2J/(R_gas*ztemp)+34.9*cal2J/R_gas)+1.0_dp)
      alpha(KPP_CO2)     = 1.0_dp/(1.0_dp+1.0_dp/(1.0_dp/(1.0_dp/1.0E-2 -1.0_dp)* &
                           EXP(2000.*ztcorr)))
      alpha(KPP_HCl)     = 1.0_dp/(EXP(-3.072E3/ztemp + 1.283E1_dp)+1.0_dp)
      !T=290: 0.096, T=270: 0.190
      !old value: alpha(KPP_ClNO3)   = 1.0E-01_dp
      alpha(KPP_ClNO3)   = 0.108_dp !mz_rs_20040308 new ref1647
      alpha(KPP_Cl2)     = 1.0_dp/(EXP(-1.3E4*cal2J/(R_gas*ztemp)+50.*cal2J/R_gas)+1.0_dp)
      alpha(KPP_HBr)     = 1.0_dp/(EXP(-3.94E3/ztemp + 1.664E1_dp) + 1.0_dp)
      !T=290K: 0.017, T=270K: 0.130
      alpha(KPP_HOBr)    = 5.0E-01_dp
      !old value: alpha(KPP_BrNO3)   = 8.0E-01_dp
      alpha(KPP_BrNO3)   = 0.063_dp !mz_rs_20040308 new ref1647
      alpha(KPP_Br2)     = 1.0_dp/(EXP(-1.3E4*cal2J/(R_gas*ztemp)+50.*cal2J/R_gas)+1.0_dp)
      alpha(KPP_BrCl)    = 0.33_dp  ! #840 alpha(KPP_Cl2)
      alpha(KPP_HOCl)    = alpha(KPP_HOBr)
      alpha(KPP_SO2)     = 1.1E-01_dp
      alpha(KPP_CH3SO3H) = 1.0_dp/(EXP(-3.50E3*cal2J/(R_gas*ztemp)+16.7*cal2J/R_gas)+1.0_dp)
      alpha(KPP_DMSO)    = 1.0_dp/(EXP(-5.12E3*cal2J/(R_gas*ztemp)+23.1*cal2J/R_gas)+1.0_dp)
      alpha(KPP_INO3)    = 1.0_dp/(1.0_dp+1.0_dp/(1.0_dp/(1.0_dp/1.0E-1 -1.0_dp)* &
                           EXP(2000.*ztcorr)))
      alpha(KPP_HOI)     = alpha(KPP_HOBr)
      alpha(KPP_HI)      = 1.0_dp/(EXP(-4.13E3/ztemp + 1.715E1_dp)+1.0_dp)
      alpha(KPP_I2)      = 1.0_dp/(1.0_dp+1.0_dp/(1.0_dp/(1.0_dp/ 1.0E-2 -1.0_dp)* &
                           EXP(2000.*ztcorr)))
      alpha(KPP_IO)      = 1.0_dp/(1.0_dp+1.0_dp/(1.0_dp/(1.0_dp/ 5.0E-1 -1.0_dp)* &
                           EXP(2000.*ztcorr)))
      alpha(KPP_I2O2)    = 1.0_dp/(1.0_dp+1.0_dp/(1.0_dp/(1.0_dp/ 1.0E-1 -1.0_dp)* &
                           EXP(2000.*ztcorr)))
      alpha(KPP_ICl)     = 1.0_dp/(1.0_dp+1.0_dp/(1.0_dp/(1.0_dp/ 1.0E-2 -1.0_dp)* &
                           EXP(2000.*ztcorr)))
      alpha(KPP_IBr)     = 1.0_dp/(1.0_dp+1.0_dp/(1.0_dp/(1.0_dp/ 1.0E-2 -1.0_dp)* &
                           EXP(2000.*ztcorr)))
      alpha(KPP_INO2)    = 1.0_dp/(1.0_dp+1.0_dp/(1.0_dp/(1.0_dp/ 1.0E-1 -1.0_dp)* &
                           EXP(2000.*ztcorr)))
      alpha(KPP_OIO)     = 0.01_dp
      alpha(KPP_HIO3)    = 0.01_dp

    END SUBROUTINE mecca_mbl_alpha

    ! ------------------------------------------------------------------------

  END SUBROUTINE mecca_mbl_trans_coeff

  ! --------------------------------------------------------------------------

  SUBROUTINE mecca_mbl_henry

      ! Temp dependent Henry constants
      ! inverse dimensionless Henry constant: k_(H,inv)^cc:=1/(k_H^cp*RT)
      ! in equilibrium: XXXaq = k_H^cp * LWC * XXXg

      ! 0.082=8.3145*10^3/101325=R/p_0*10^3 :
      ! R includes conversion from M/atm --> mol/(m^3*Pa)
      ! so k_H^cp is taken as M/atm (the most common literature unit): 
      ! SEE END OF SR

      INTEGER  :: jn
      REAL(dp) :: zhenry(0:nspec), zfac

      zhenry(:)   = -999.999_dp !qqq dummy value

      zhenry(KPP_H2SO4)  = 1.E+11_dp ! NIST --> RS_Gmitro_Vermeulen
      zhenry(KPP_HI)     = 0.0_dp       ! (dissociation)

      ! explicitly Temp dependent

      zhenry(KPP_NO)     = henry_func(1.9E-03_dp,         1480._dp) !RS_Lide
      zhenry(KPP_NO2)    = henry_func(6.4E-03_dp,         2500._dp) !RS_MOCCA
      zhenry(KPP_HNO3)   = henry_func(2.5E6_dp/1.5E1_dp,  8694._dp) !RS_MOCCA_exakt
      zhenry(KPP_HNO4)   = henry_func(1.2E4_dp,           6900._dp) !
      zhenry(KPP_NH3)    = henry_func(58._dp,             4085._dp) !RS_MOCCA
      zhenry(KPP_SO2)    = henry_func(1.2E0_dp,           3120._dp) !RS_MOCCA
      zhenry(KPP_O3)     = henry_func(1.2E-02_dp,         2560._dp) !RS_MOCCA
      zhenry(KPP_HCOOH)  = henry_func(3.7E+03_dp,         5700._dp) !RS_MOCCA
      zhenry(KPP_HCHO)   = henry_func(7.0E+03_dp,         6425._dp) !RS_MOCCA
      zhenry(KPP_H2O2)   = henry_func(1.E+05_dp,          6338._dp) !RS_MOCCA
      zhenry(KPP_CH3OOH) = henry_func(3.0E+02_dp,         5322._dp) !RS_MOCCA CH3OOH
      zhenry(KPP_HONO)   = henry_func(4.9E+01_dp,         4780._dp) !RS_MOCCA
      zhenry(KPP_HCl)    = henry_func(2._dp/1.7E0_dp,     9001._dp) !RS_MOCCA_exakt
      zhenry(KPP_NO3)    = henry_func(2._dp,              2000._dp) !RS_MOCCA
      zhenry(KPP_DMSO)   = henry_func(5.E4_dp,            6425._dp) !RS_MOCCA 
      zhenry(KPP_HOCl)   = henry_func(6.7E2_dp,           5862._dp) !RS_MOCCA
      zhenry(KPP_Cl2)    = henry_func(9.1E-2_dp,          2500._dp) !RS_MOCCA 
      zhenry(KPP_HBr)    = henry_func(1.3E0_dp,          10239._dp) !RS_MOCCA
      zhenry(KPP_Br2)    = henry_func(7.6E-1_dp,          4094._dp) !RS_MOCCA
      zhenry(KPP_BrCl)   = henry_func(9.4E-1_dp,          5600._dp) !RS_MOCCA
      zhenry(KPP_HOBr)   = henry_func(9.3E1_dp,           5862._dp) !RS_MOCCA 
      zhenry(KPP_I2)     = henry_func(3._dp,              4431._dp) !RS_MOCCA
      zhenry(KPP_HOI)    = henry_func(4.5E2_dp,           5862._dp) !RS_MOCCA 
      zhenry(KPP_ICl)    = henry_func(1.1E2_dp,           5600._dp) !RS_MOCCA 
      zhenry(KPP_IBr)    = henry_func(2.4E1_dp,           5600._dp) !RS_MOCCA 
      zhenry(KPP_OH)     = henry_func(3.0E1_dp,           4300._dp) !RS_MOCCA 
      zhenry(KPP_HO2)    = henry_func(3.9E3_dp,           5900._dp) !RS_MOCCA 
      zhenry(KPP_CH3O2)  = henry_func(6._dp,              5600._dp) !RS_Jacob
      zhenry(KPP_IO)     = henry_func(4.5E2_dp,           5862._dp) !RS_MOCCA 
      zhenry(KPP_CO2)    = henry_func(3.1E-02_dp,         2423._dp) !RS_MOCCA
      zhenry(KPP_O2)     = henry_func(1.3E-3_dp,          1500._dp) !RS_Lide95 

      ! dummy henry constants
      ! mz_ak_20030514    zhenry(KPP_DGtAsg) = 3.E11
      ! mz_ak_20030514    zhenry(KPP_DGtAig) = 1.

      ! unit: mol/(l*atm) --> mol(aq)/m3(aq) / mol(g)/m3(g) (i.e. dimensionless)
      ! FCT=1.e3*8.3145*T/p_0=0.082*T
      ! PLUS conversion to inverse henry constant:  h_(H,inv)^cc = 1/k_H^cc
      ! i.e. mol(aq)/m3(aq) / mol(g)/m3(air) ->  mol(g)/m3(air) / mol(aq)/m3(aq)

      zfac = R_gas / 101325. * 1.E3 * temp

      yhenry(:)  = 0.
      DO jn=1,nspec
        IF (zhenry(jn).GT.0._dp) THEN
          yhenry(jn) = 1./(zhenry(jn)*zfac)
        ELSE
          ! "else": zhenry=0 <=> k_H^cc=infinity
          yhenry(jn) = 0._dp
        ENDIF
      ENDDO

    END SUBROUTINE mecca_mbl_henry

    ! ------------------------------------------------------------------------

    REAL(dp) FUNCTION henry_func(pa, pb)

      REAL(dp) :: pa, pb

      INTRINSIC EXP

      ! 1/298.=3.3557e-3
      henry_func = pa*EXP(pb*((1._dp/temp)-3.3557E-3_dp))

    END FUNCTION henry_func

    ! ------------------------------------------------------------------------

    SUBROUTINE mecca_mbl_calc_k_ex
      
    ! calculation of gas-aqueous-phase exchange coefficients   

      INTEGER jn, zkc
      REAL :: zhetT

      k_exf(:,:) = 0.
      k_exb(:,:) = 0.
      DO jn = 1,nspec   ! loop over species
        DO zkc = AS, CS ! loop over modes
          ! reaction rate coefficients
          k_exf(zkc,jn) = xaer(zkc)*ykmt(zkc,jn)*lwc(zkc)   ! forward
          k_exb(zkc,jn) = xaer(zkc)*ykmt(zkc,jn)*yhenry(jn) ! backward
        ENDDO
      ENDDO

      zhetT = C(KPP_H2O_as)
      IF (KPP_Clm_as/=0) zhetT = zhetT + 5.E2 * C(KPP_Clm_as)
      IF (KPP_Brm_as/=0) zhetT = zhetT + 3.E5 * C(KPP_Brm_as)
      k_exf_N2O5(AS)  = xaer(AS) * lwc(AS) * ykmt(AS,KPP_N2O5)  / zhetT
      k_exf_ClNO3(AS) = xaer(AS) * lwc(AS) * ykmt(AS,KPP_ClNO3) / zhetT
      k_exf_BrNO3(AS) = xaer(AS) * lwc(AS) * ykmt(AS,KPP_BrNO3) / zhetT

      zhetT = C(KPP_H2O_cs)
      IF (KPP_Clm_cs/=0) zhetT = zhetT + 5.E2 * C(KPP_Clm_cs)
      IF (KPP_Brm_cs/=0) zhetT = zhetT + 3.E5 * C(KPP_Brm_cs)
      k_exf_N2O5(CS)  = xaer(CS) * lwc(CS) * ykmt(CS,KPP_N2O5)  / zhetT
      k_exf_ClNO3(CS) = xaer(CS) * lwc(CS) * ykmt(CS,KPP_ClNO3) / zhetT
      k_exf_BrNO3(CS) = xaer(CS) * lwc(CS) * ykmt(CS,KPP_BrNO3) / zhetT

    END SUBROUTINE mecca_mbl_calc_k_ex

  ! --------------------------------------------------------------------------

    REAL(dp) FUNCTION hrate_diff(pmode,pgspecies,paspecies)

      INTEGER  :: pmode, pgspecies, paspecies
      REAL(dp) :: forward, backward, r_xaer

      INTRINSIC REAL

      r_xaer = REAL(xaer(pmode),dp)
      forward  = r_xaer*ykmt(pmode,pgspecies)*lwc(pmode)        * C(pgspecies)
      backward = r_xaer*ykmt(pmode,pgspecies)*yhenry(pgspecies) * C(paspecies)
      hrate_diff = (backward - forward) * R_gas * temp / &
        ( (N_A/1.E6) * press ) * 86400.

    END FUNCTION hrate_diff
!----------------------------------------------------------------------------

elemental real(dp) function density(press, temp, sphum)

  ! calculate air density in kg/m3

  USE messy_main_constants_mem, ONLY: R_gas, M_air, M_H2O

  REAL(dp), INTENT(in) :: press
  REAL(dp), INTENT(in) :: temp
  REAL(dp), INTENT(in) :: sphum

  density = press /( R_gas * &
       temp * (1._dp +M_air / M_H2O -1._dp *sphum))

end function density

!----------------------------------------------------------------------------
!-----------------------------------------------------------------------------

elemental real(dp) function layerthickness(geopot_u, geopot_l)

  USE messy_main_constants_mem,  ONLY: g
  
  REAL(dp), INTENT(in) :: geopot_u
  REAL(dp), INTENT(in) :: geopot_l
  !LOCAL 

     layerthickness  = (geopot_u-geopot_l)/ g
   end function layerthickness

!-----------------------------------------------------------------------------

!*****************************************************************************

END MODULE messy_mecca_mbl

!*****************************************************************************
