# ----------------------------------------------------------------------------
#
# Authors:
#   Rolf Sander, Max-Planck-Institute, Mainz, Germany, 2003-2004
#   Astrid Kerkweg, Max-Planck-Institute, Mainz, Germany, 2004
#
# Time-stamp: <2004-07-15 16:27:29 sander>
#
# eqn2tex.awk transforms eqn into tex.
#
# usage:
# gawk -f eqn2tex.awk ../mecca.eqn
#
# notes:
#
# - lines starting with "#DEBUG" can be activated for debugging
#
# ----------------------------------------------------------------------------

BEGIN {
  # remove "../", then replace "." by "_" in input file
  basename = gensub("\\.\\./", "", "g", ARGV[1])
  gsub("\\.", "_", basename)
  # define name of output files
  hfile   = basename "_h.tex"
  afile   = basename "_a.tex"
  gfile   = basename "_g.tex"
  jfile   = basename "_j.tex"
  pscfile = basename "_psc.tex"
  eqfile  = basename "_eq.tex"
  # write header line
  dontedit = "% this file was created automatically by eqn2tex, do not edit!"
  print dontedit > hfile
  print dontedit > afile
  print dontedit > gfile
  print dontedit > jfile
  print dontedit > pscfile
  print dontedit > eqfile
  # initialize some strings
  errorstring = ""
  unknown = "????"
}

# ----------------------------------------------------------------------------

{
  # store reaction number like {#3103b} in number
  if (match($0, "{#([^}]*)}", arr) != 0) {
    number = arr[1]
    #DEBUG printf "number   = %s\n", number
  } else {
    number = unknown
    #DEBUG printf "no number found\n"
  }
  # does current line contain a chemical equation, i.e. something like
  # " = ... : ... ; " ?
  if (match($0, "=.*:.*;") != 0) {
    #DEBUG printf "equation = %s\n", $0
    # check if reaction number exists
    if (number==unknown) {
      errorstring = sprintf("%s\nERROR: This reaction has no number:\n  %s",
        errorstring, $0)
    }
    # store marker like {%StTrGNJ} in marker
    if (match($0, "{%([^}]*)}", arr) != 0) {
      marker = arr[1]
      #DEBUG printf "marker   = %s\n", marker
    } else {
      marker = ""
      #DEBUG printf "no marker found\n"
    }
    # store alternative LaTeX text for the rate
    # constant's A-factor like {@} in klatex
    if (match($0, "{@([^}]*)}", arr) != 0) {
      klatex = arr[1]
      # replace <> by {}
      gsub("<", "{", klatex)
      gsub(">", "}", klatex)
      #DEBUG printf "klatex   = %s\n", klatex
    } else {
      klatex = ""
      #DEBUG printf "no klatex found\n"
    }
    # store alternative LaTeX text for the rate constant's exponent
    # like {$} in kexplatex
    if (match($0, "{\\$([^}]*)}", arr) != 0) {
      kexplatex = arr[1]
      # replace <> by {}
      gsub("<", "{", kexplatex)
      gsub(">", "}", kexplatex)
      #DEBUG printf "kexplatex   = %s\n", kexplatex
    } else {
      kexplatex = ""
      #DEBUG printf "no kexplatex found\n"
    }
    # store BibTeX references like {&1400} in ref
    if (match($0, "{(&+)([^}]*)}", arr) != 0) {
      ref = arr[2]
      if (ref=="") {
        ref = "see note"
      } else {
        # put each BibTeX reference into \citet{}
        # here, "&" represents the matched regexp
        gsub("[^, ]+", "\\citet{&}", ref)
        # if there are two ampersands like {&&1400} read also the footnote
        if (arr[1]=="&&") {
          ref = ref "$^*$"
        }
      }
      #DEBUG printf "ref      = %s\n", ref
    } else {
      if (match(number, "^EQ[0-9]*b") != 0) {
        # equilibrium backward reaction EQ##b doesn't need a reference
        #DEBUG printf "no ref needed for reaction %s\n", number
      } else {
        ref = "\\rule{8mm}{3mm}"
        errorstring = sprintf("%s\nERROR: reaction %s has no reference",
          errorstring, number)
        #DEBUG printf "no ref found\n"
      }
    }
    #DEBUG printf "equation = %s\n", $0
    # delete all comments {...} from $0
    gsub("{[^}]*}", "")
    #DEBUG printf "equation = %s\n", $0
    # reduce multiple spaces to one
    gsub("  +", " ")
    #DEBUG printf "equation = %s\n", $0
    # split into equation and rateconst using the kpp-syntax separators ":;"
    split($0, arr, " *[:;] *")
    #DEBUG printf "equation  = %s\n", equation
    #DEBUG printf "rateconst = %s\n", rateconst
    equation  = arr[1]
    rateconst = "\\code{" arr[2] "}"
    # put LaTeX command \kpp{} around each specie in equation
    # here, "&" represents the matched regexp
    gsub("[A-Za-z][A-Za-z0-9_]*", "\\kpp{&}", equation)
    #DEBUG printf "equation  = %s\n", equation
    # create reaction arrow
    gsub("=", "$\\rightarrow$", equation)
    #DEBUG printf "equation  = %s\n", equation
    # use alternative LaTeX text for rate constant base?
    if (klatex!="") {
      rateconst = klatex
    }
    # write a line into the specific LaTeX table
    if (match(number, "^G[0-9]") != 0) {
      # G = gas-phase reaction
      printf "%s & %s & %s & %s & %s\\\\\n",
        number, marker, equation, rateconst, ref >> gfile
      if (kexplatex!="") {
        errorstring = sprintf("%s\nERROR: reaction %s cannot use {$%s}",
          errorstring, number, kexplatex)
      }
    } else if (match(number, "^J[0-9]") != 0) {
      # J = photolysis reaction
      printf "%s & %s & %s & %s & %s\\\\\n",
        number, marker, equation, rateconst, ref >> jfile
      if (kexplatex!="") {
        errorstring = sprintf("%s\nERROR: reaction %s cannot use {$%s}",
          errorstring, number, kexplatex)
      }
    } else if (match(number, "^PSC[0-9]") != 0) {
      # PSC = PSC reaction
      printf "%s & %s & %s & %s & %s\\\\\n",
        number, marker, equation, rateconst, ref >> pscfile
      if (kexplatex!="") {
        errorstring = sprintf("%s\nERROR: reaction %s cannot use {$%s}",
          errorstring, number, kexplatex)
      }
    } else if (match(number, "^A[0-9]") != 0) {
      # A = aqueous-phase reaction
      printf "%s & %s & %s & %s & %s & %s\\\\\n",
        number, marker, equation, rateconst, kexplatex, ref >> afile
    } else if (match(number, "^EQ[0-9]*b") != 0) {
      # backward reaction, do nothing
    } else if (match(number, "^EQ[0-9]*f") != 0) {
      # EQ = equilibria (acid-base and others)
      # delete f in equation number
      match(number, "EQ[0-9]*", eqnumber)
      # replace rightarrow by rightleftharpoons
      gsub("rightarrow", "rightleftharpoons", equation)
      printf "%s & %s & %s & %s & %s & %s\\\\\n",
        eqnumber[0], marker, equation, rateconst, kexplatex, ref >> eqfile
    } else if (match(number, "^H[0-9]") != 0) {
      # H = aqueous-phase heterogenous and Henry reaction
      printf "%s & %s & %s & %s & %s\\\\\n",
        number, marker, equation, rateconst,  ref >> hfile
      if (kexplatex!="") {
        errorstring = sprintf("%s\nERROR: reaction %s cannot use {$%s}",
          errorstring, number, kexplatex)
      }
    } else {
      errorstring = sprintf("%s\nERROR: unknown type of reaction number: %s",
        errorstring, number)
      #DEBUG printf "%s & %s & %s & %s & %s & %s\\\\\n", number, marker, equation, rateconst, kexplatex, ref
    }
    #DEBUG printf "\n"
  } else {
    # not an equation line
    # check for LaTeX text {@...}
    if (match($0, "{@([^}]*)}", arr) != 0) {
      klatex = arr[1]
      #DEBUG printf "klatex   = %s\n", klatex
      # write a line into the specific LaTeX table
      if (match(number, "^G[0-9]") != 0) {
        # G = gas-phase reaction
        printf "%s\n", klatex >> gfile
      } else if (match(number, "^J[0-9]") != 0) {
        # J = photolysis reaction
        printf "%s\n", klatex >> jfile
      } else if (match(number, "^PSC[0-9]") != 0) {
        # PSC = PSC reaction
        printf "%s\n", klatex >> pscfile
      } else if (match(number, "^A[0-9]") != 0) {
        # A = aqueous-phase reaction
        printf "%s\n", klatex >> afile
      } else if (match(number, "^EQ[0-9]") != 0) {
        # EQ = equilibria (acid-base and others)
        printf "%s\n", klatex >> eqfile
      } else if (match(number, "^H[0-9]") != 0) {
        # H = aqueous-phase heterogenous and Henry reaction
        printf "%s\n", klatex >> hfile
      } else {
        errorstring = sprintf("%s\nERROR: unknown type of reaction number: %s",
          errorstring, number)
      }
    }
  }
}

# ----------------------------------------------------------------------------

END {
  print errorstring
}

# ----------------------------------------------------------------------------
