# ----------------------------------------------------------------------------

# Author: Rolf Sander, Max-Planck-Institute, Mainz, Germany, 2003
# Time-stamp: <2004-04-02 16:38:42 sander>

# tracdef.awk extracts all KPP_XXX from messy_mecca_kpp_g_mem.f90 and
# inserts these species into messy_mecca_idt_e5.inc. For all KPP_XXX that are
# not zero it also inserts the species into messy_mecca_c2mr_e5.inc,
# messy_mecca_mr2c_e5.inc, and messy_mecca_trac_e5.inc, using the tracer info
# in tracdef.tex. For the box model version, it also creates species.inc.

# 1) messy_mecca_idt_e5.inc:  the index declaration "INTEGER idt_XXX"
# 2) messy_mecca_c2mr_e5.inc: the conversion concentration -> mixing ratio
# 3) messy_mecca_mr2c_e5.inc: the conversion mixing ratio -> concentration
# 4) messy_mecca_trac_e5.inc: the "CALL new_tracer" command including all tracer
#                             properties as defined in tracdef.tex.
# 5) species.inc:             string definitions of species names

# usage:
# gawk -f tracdef.awk -v tracdef=tracdef.tex messy_mecca_kpp_g_mem.f90

# Normally, however, tracdef.awk is called by xmecca and not started
# directly from the command line.

# ----------------------------------------------------------------------------

BEGIN {
# define names of output files
idtfile     = "messy_mecca_idt_e5.inc"
conc2mrfile = "messy_mecca_c2mr_e5.inc"
mr2concfile = "messy_mecca_mr2c_e5.inc"
tracfile    = "messy_mecca_trac_e5.inc"
speciesfile = "species.inc"
# write header line
dontedit = "! -*- f90 -*- this file was created by xmecca, do not edit!"
print dontedit > idtfile
print dontedit > conc2mrfile
print dontedit > mr2concfile
print dontedit > tracfile
print dontedit > speciesfile
print "List of tracers:"
errorstring = ""
}

# ----------------------------------------------------------------------------

# if prop[i] contains more than just spaces, add an optional parameter
# to "CALL new_tracer" in tracfile. Then, increment i so that the next
# property is check in the folowing call of addproperty

function addproperty(property)
 {
  if (match(prop[i], "^ *$") == 0) 
    printf "  %s = %s, &\n", property, prop[i] >> tracfile
  i++
  }

# # mz_ak_20031127+ 
# {
#   if (match(property, "nrerun" ) == 0 && match(property, "force") == 0) { 
#     if (match(prop[i], "^ *$") == 0) 
#       printf "  %s = %s, &\n", property, prop[i] >> tracfile
# 	} else{
# 	  printf "  %s = OFF, &\n", property, prop[i] >> tracfile
# 	    }     
#   i++
#   }
# mz_ak_20031127- 

# ----------------------------------------------------------------------------

{
# is current line something like "INTEGER ... KPP_XXX = nnn" with any nnn ?
if (match($0, "INTEGER.*KPP_([A-Za-z0-9_]+) = ", arr) != 0) {
  # KPP_XXX was found and the species name XXX, i.e. the "([A-Za-z0-9]+)" part
  # of the above regexp, is stored in arr[1].
  # define hfill for vertical alignment
  hfill = substr("            ", 1, 12-length(arr[1]))
  # 1) add tracer to messy_mecca_idt_e5.inc
  printf "INTEGER :: idt_%s%s= 0\n", arr[1], hfill >> idtfile
}
# is current line something like "INTEGER ... KPP_XXX = nnn" with nnn != 0 ?
if (match($0, "INTEGER.*KPP_([A-Za-z0-9_]+) = [^0]", arr) != 0) {
  # split arr (field separator is "_" ) and save basename
  # and subname in array basesub
  split(arr[1], basesub, "_")
  # Now try to find this species in the tracdef file. The species name
  # must be the first non-whitespace word in the line and enclosed in
  # single quotation marks: 'XXX'.
  if (basesub[2]=="") {
    subhfill = ""
    command = ("grep -E \"^[ 	]*'" basesub[1] "'\" " tracdef)
  } else {
    subhfill = substr("            ", 1, 1+length(basesub[2]))
    command = ("grep -E \"^[ 	]*'" basesub[1] "'[ 	]*&[ 	]*'" basesub[2] "'\" " tracdef)
  }
  if ((command | getline) > 0) {
    # 'XXX' was found and stored in $0 by getline
    printf "%s ", arr[1]
    # ------------------------------------------------------------------------
    # 2) add tracer to messy_mecca_c2mr_e5.inc
    printf "zmrac(idt_%s)%s= riac * C(KPP_%s)\n",
      arr[1], hfill, arr[1] >> conc2mrfile
    # ------------------------------------------------------------------------
    # 3) add tracer to messy_mecca_mr2c_e5.inc
    printf "C(KPP_%s)%s= cair * zmrbc(idt_%s)\n",
      arr[1], hfill, arr[1] >> mr2concfile
    # ------------------------------------------------------------------------
    # 4) add tracer to messy_mecca_trac_e5.inc
    # start of CALL new_tracer
    printf "%s%s%s%s%s%s\n", "CALL new_tracer(status, setname, '", 
      basesub[1], "',", hfill, subhfill, "modstr, &" >> tracfile
    printf "  quantity             = AMOUNTFRACTION, &\n" >> tracfile
    printf "  unit                 = 'mol/mol', &\n"      >> tracfile
    # optional parameters for new_tracer if defined in the tracdef file.
    # split $0 (field separator is & or \\ surrounded by spaces)
    # save individual tracer properties in elements of array prop
    split($0, prop, " *& *| *\\\\")
    # prop[1] is the tracer name, optional properties start at i=2
    i=2
    # the order of the addproperty calls must be the same as the order of
    # the corresponding entries in the tracdef file.
    addproperty("subname             ")
    addproperty("medium              ")
    addproperty("lrestart            ")
    addproperty("nforce_col          ")
    addproperty("vini                ")
    addproperty("molarmass           ")
    addproperty("ndrydep             ")
    addproperty("henry               ")
    addproperty("dryreac_sf          ")
    addproperty("lforce_init         ")
    addproperty("nadvect             ")
    addproperty("nvdiff              ")
    addproperty("nconvect            ")
    addproperty("nwetdep             ")
    addproperty("nsedi               ")
    addproperty("nscav               ")
    addproperty("nintegrate          ")
    addproperty("m_aerosol_modelname ")
    addproperty("m_aerosol_mode      ")
    addproperty("lav                 ")
    addproperty("lsd                 ")
    # last part of CALL new_tracer
    printf "  idx                  = idt_%s)\n", arr[1] >> tracfile
    printf "CALL tracer_halt(substr,status)\n" >> tracfile
    # ------------------------------------------------------------------------
  } else {
    # print current species in parenthesis
    printf "(%s) ", arr[1]
    # add another line to errorstring
    if (arr[1]=="H2O") {
      # Missing H2O is okay. This tracer is not provided by MECCA.
      errorstring = sprintf("%s\nINFO:    %s%s not found in %s (that's okay)", 
        errorstring, arr[1], hfill, tracdef)
    } else {
      # Other species should not be missing in tracdef.tex.
      errorstring = sprintf("%s\nWARNING: %s%s not found in %s", 
        errorstring, arr[1], hfill, tracdef)
    }
  }
  # --------------------------------------------------------------------------
  # 5) add tracer name to species.inc
  printf "species(kpp_%s)%s= '%s'\n",
    arr[1], hfill, arr[1] >> speciesfile
  # close grep command to avoid too many open processes
  close(command)
}
}

# ----------------------------------------------------------------------------

END {
print errorstring
printf "(warnings about %s are only for ECHAM, not for the box model)\n\n",
  tracdef
}

# ----------------------------------------------------------------------------
