/***********************************************************************************
  KPP - symbolic chemistry Kinetics PreProcessor

  developed at Center for Global and Regional Atmospheric Research,
               University of Iowa

            by Valeriu Damian & Adrian Sandu 

***********************************************************************************/

#define MIN(b,c) ((b) <  (c) ) ?(b):(c)

#include "gdata.h"
#include "code.h"
#include <string.h>

#define MAX_LINE 120

char *F95_types[] = { "",               /* VOID */ 	/* F95 Done */
                    "INTEGER",          /* INT */
                    "REAL",             /* FLOAT */
                    "REAL(dp)",         /* DOUBLE */ /* mz_rs_20040202 */
                    "CHARACTER*12"      /* STRING */
                  };

void F95_WriteElm( NODE * n ) 			/* F95 done */
{
ELEMENT *elm;
char * name;
char maxi[20];
char maxj[20];

  elm = n->elm;
  name = varTable[ elm->var ]->name;

  switch( n->type ) {
    case CONST: bprintf("%g", elm->val.cnst);
		break;
    case ELM:   bprintf("%s", name);
		break;
    case VELM:  if( elm->val.idx.i >= 0 ) sprintf( maxi, "%d", elm->val.idx.i+1 ); /*changed here */
                  else sprintf( maxi, "%s", varTable[ -elm->val.idx.i ]->name );
//                bprintf("DIMENSION(%s) :: %s", maxi, name );
		bprintf("%s(%s)", name, maxi );
		break;
    case MELM:  if( elm->val.idx.i >= 0 ) sprintf( maxi, "%d", elm->val.idx.i+1 );
                  else sprintf( maxi, "%s", varTable[ -elm->val.idx.i ]->name );
                if( elm->val.idx.j >= 0 ) sprintf( maxj, "%d", elm->val.idx.j+1 );
                  else sprintf( maxj, "%s", varTable[ -elm->val.idx.j ]->name ); /* changed here */
                  bprintf("%s(%s,%s)", name, maxi, maxj );
//                bprintf("DIMENSION(%s,%s) :: %s", maxi, maxj, name );
		break;
    case EELM:  bprintf("(%s)", elm->val.expr ); /* no change */
		break;
  }
}

void F95_WriteSymbol( int op ) 			/* F95 done, no change */	
{
  switch( op ) {
    case ADD:   bprintf("+"); 
                AllowBreak();
    		break;
    case SUB:   bprintf("-"); 
                AllowBreak();
    		break;
    case MUL:   bprintf("*"); 
                AllowBreak();
    		break;
    case DIV:   bprintf("/"); 
                AllowBreak();
    		break;
    case POW:   bprintf("power");
                break;		
    case O_PAREN: bprintf("(");
                AllowBreak();
                break;            
    case C_PAREN: bprintf(")");
                break;            
    case NONE:
                break;            
  }
}

void F95_WriteAssign( char *ls, char *rs )	
{
int start;
int linelg;
int i;
char c;
int first;
int crtident;
  
  crtident = 6 + ident * 2;			
  bprintf("%*s%s = ", crtident, "", ls);
  start = strlen( ls ) + 2;			
  linelg = 70 - crtident - start - 1;

  first = 1;
  while( strlen(rs) > linelg ) {
    for( i=linelg; i>10; i-- )
      if( rs[i] & 0x80 )
        break;
    if( i == 10 ) { /* mz_rs_20020430: i == 10 if no 0x80 was found */
      /* now try to break line at "+" or "*" */
      for( i=linelg; i>10; i-- )
        if ( (rs[i]=='+') || (rs[i]=='*') || (rs[i]==',') )
          break;
      if( i == 10 ) { /* mz_rs_20020430: i == 10 if no "+" or "*" was found */
        /* printf("\nPossible error when cutting lines:\n"); */
        printf("\nLine break may not be nice:\n"); /* mz_rs_20030305 */
        printf("%s\n", rs ); /* mz_rs_20020920: more info for user */
        i = linelg;
      }
    }
    c = rs[i]; 
    rs[i] = 0;
    if ( first ) { 
      bprintf("%s", rs );
      linelg++;
      first = 0;
    } else {
      /* the next line will be continued with at least 2 more lines */
      /* mz_rs_20030305: "&" in continuation line added */
      bprintf("&\n     %*s&%s", start, "", rs );			/* format changed here */
    }  
    rs[i] = c;
    rs += i;
  }
  
  if ( first ) bprintf("%s\n", rs );
          /* the last continuation line will follow after this line */
          /* mz_rs_20030305: "&" in continuation line added */
          else bprintf("&\n     %*s&%s\n", start, "", rs );		/* format changed here */

  FlushBuf(); 
}



void F95_WriteComment( char *fmt, ... )		/* F95 done */
{
Va_list args;
char buf[ MAX_LINE ];

  Va_start( args, fmt );
  vsprintf( buf, fmt, args );
  va_end( args );
  bprintf( "! %-65s\n", buf );			/* changed here */

  FlushBuf();
}

char * F95_Decl( int v )			
{
static char buf[120];
VARIABLE *var;
char *baseType;
char maxi[20];
char maxj[20];

  var = varTable[ v ];
  baseType = F95_types[ var->baseType ];
  
  *buf = 0;

  switch( var->type ) {				/* changed here */
    case ELM:   sprintf( buf, "%s :: %s",
                        baseType, var->name );
		break;
    case VELM:  
                if( var->maxi > 0 ) sprintf( maxi, "%d", var->maxi );
                  else sprintf( maxi, "%s", varTable[ -var->maxi ]->name );  
                if( (var->maxi == 0) || 
                    ((var->maxi < 0) && (varTable[ -var->maxi ]->maxi == 0)) )
                  strcat( maxi, "+1");
                sprintf( buf, "%s, DIMENSION(%s) :: %s",	/* changed here */
                        baseType, maxi, var->name );
		break;
    case MELM:  if( var->maxi > 0 ) sprintf( maxi, "%d", var->maxi );
                  else sprintf( maxi, "%s", varTable[ -var->maxi ]->name );  
                if( (var->maxi == 0) || 
                    ((var->maxi < 0) && (varTable[ -var->maxi ]->maxi == 0)) )
                  strcat( maxi, "+1");
                if( var->maxj > 0 ) sprintf( maxj, "%d", var->maxj );
                  else sprintf( maxj, "%s", varTable[ -var->maxj ]->name );  
                if( (var->maxj == 0) || 
                    ((var->maxj < 0 ) && (varTable[ -var->maxj ]->maxi == 0)) )
                  strcat( maxj, "+1");
                sprintf( buf, "%s, DIMENSION(%s,%s) :: %s",	/* changed here */		
                         baseType, maxi, maxj,var->name );  
		break;
    default:
                printf( "Can not declare type %d", var->type );
                break;
  }
  return buf;
}

void F95_Declare( int v )		/* no change */
{
  if( varTable[ v ]->comment ) {
    F95_WriteComment( "%s - %s", 
                    varTable[ v ]->name, varTable[ v ]->comment );
  }
  bprintf("      %s\n", F95_Decl(v) );	/* format changed here */

  FlushBuf();
}

void F95_ExternDeclare( int v )		/* no change */
{
  F95_Declare( v );
//  bprintf("      COMMON /%s/ %s\n", CommonName, varTable[ v ]->name );	/* F95 second */
}

void F95_GlobalDeclare( int v )		/* no change */
{
}

void F95_DeclareConstant( int v, char *val )	/* F95 done */ 
{
VARIABLE *var;
int ival;
char dummy_val[100];           /* used just to avoid strange behaviour of
                                  sscanf when compiled with gcc */
                                  
  strcpy(dummy_val,val);val = dummy_val;

  var = varTable[ v ];
  
  if( sscanf(val, "%d", &ival) == 1 )
    if( ival == 0 ) var->maxi = 0;
               else var->maxi = 1;
  else
    var->maxi = -1;       
  
  if( var->comment ) 
    F95_WriteComment( "%s - %s", var->name, var->comment );

  switch( var->type ) {
	  case CONST: bprintf("      %s, PARAMETER :: %s = %s \n",	/* F95 line changed here */
                       F95_types[ var->baseType ], var->name, val );
                break;       
    default:
                printf( "Invalid constant %d", var->type );
                break;
  }

  FlushBuf();
}

void F95_WriteVecData( VARIABLE * var, int min, int max, int split )	
{
char buf[80];
char *p;

  if( split )
    sprintf( buf, "%6sdata( %s(i), i = %d, %d ) / &\n%5s", 		/* F95 line format changed here */
                " ", var->name, min, max, " " );
  else
    sprintf( buf, "%6sdata %s / &\n%5s",
                    " ", var->name, " " );
                                      
  FlushThisBuf( buf );
  bprintf( " / \n\n" );
  FlushBuf();        
}

void F95_DeclareData( int v, int * values, int n )
{                                                                                
int i, j;
int nlines, min, max;
int split;
VARIABLE *var;
int * ival;
double * dval;
char **cval;
int maxCols = MAX_COLS;

  var = varTable[ v ];
  ival = (int*) values;
  dval = (double*) values;
  cval = (char**) values;
    
  nlines = 1;
  min = max = 1;
  split = 0;

  switch( var->type ) {
    case VELM: if( n <= 0 ) break;
    	       for( i = 0; i < n; i++ ) {
                 switch( var->baseType ) {
                   case INT: bprintf( "%3d",  ival[i] ); maxCols=12; break;
                   case DOUBLE: 
                   case REAL:bprintf( "%5lg", dval[i] ); maxCols=8; break;
                   case STRING:bprintf( "'%s'", cval[i] ); maxCols=5; break;
                 }
                 if( ( (i+1) % 12 == 0 ) && ( nlines > MAX_LINES ) ) { 		/* no change here */
                     split = 1; nlines = 1;
                     F95_WriteVecData( var, min, max, split );  
                     min = max + 1;
                 } 
                 else { 
                   if( i < n-1 ) bprintf( "," );
                   if( (i+1) % maxCols == 0 ) { 
                     bprintf( " &\n%5s", " " );				/* F95 line format changed here */
                     nlines++;                 
                   }  
                 }  
                 max ++;
               }
               F95_WriteVecData( var, min, max-1, split );
               break;
                                                                 
    case ELM:  bprintf( "%6sdata %s / ", " ", var->name );		/* F95 line format changed here */
               switch( var->baseType ) {
                 case INT: bprintf( "%d",  *ival ); break;
                 case DOUBLE: 
                 case REAL:bprintf( "%lg", *dval ); break;
                 case STRING:bprintf( "'%s'", *cval ); break;
               }
               bprintf( " / \n" );
               FlushBuf();
               break;
    default:
               printf( "\n Function not defined !\n" );
               break;
  }
}

void F95_InitDeclare( int v, int n, void * values )	/* no change */
{
int i;
VARIABLE * var;

  var = varTable[ v ];
  var->maxi = max( n, 1 );

  NewLines(1);  
  F95_DeclareData( v, values, n );
}

void F95_FunctionStart( int f, int *vars )	/* no change */
{
int i;
int v;
char * name;
int narg;

  name = varTable[ f ]->name;
  narg = varTable[ f ]->maxi;

  bprintf("      SUBROUTINE %s ( ", name );
  for( i = 0; i < narg-1; i++ ) {
    v = vars[ i ];
    bprintf("%s, ", varTable[ v ]->name );
  }
  if( narg >= 1 ) {
    v = vars[ narg-1 ];
    bprintf("%s ", varTable[ v ]->name );
  }
  bprintf(")\n");

  FlushBuf();
}                  

void F95_FunctionPrototipe( int f, ... )	/* no change */
{
char * name;
int narg;

  name = varTable[ f ]->name;
  narg = varTable[ f ]->maxi;

  bprintf("      EXTERNAL %s\n", name );

  FlushBuf();
}

void F95_FunctionBegin( int f, ... )		/* no change */
{
Va_list args;
int i;
int v;
int vars[20];
char * name;
int narg;
FILE *oldf;
int SOLVE;

  name = varTable[ f ]->name;
  narg = varTable[ f ]->maxi;
    
  Va_start( args, f );
  for( i = 0; i < narg; i++ ) 
    vars[ i ] = va_arg( args, int );
  va_end( args );
    
  CommentFncBegin( f, vars );
  F95_FunctionStart( f, vars );
  NewLines(1);
//  bprintf("      INCLUDE '%s.h'\n\n", rootFileName ); /* F95 second */
  /* mz_rs_20020530: next two lines deleted because for SUBROUTINE SOLVE, */
  /* F95_FunctionBeginSprs is called instead of F95_FunctionBegin   */
  /*  if ( f == SOLVE ) */
  /* bprintf("      USE SPRSDATA\n"); */
  /* bprintf("/\* mz_rs_      USE GDATA\n"); */ /* mz_rs_20040202: deleted */

  FlushBuf();

  for( i = 0; i < narg; i++ ) 
    F95_Declare( vars[ i ] );

  bprintf("\n");
  FlushBuf();

  MapFunctionComment( f, vars );
}


void F95_FunctionBeginSprs( int f, ... )		/* no change */
{
Va_list args;
int i;
int v;
int vars[20];
char * name;
int narg;
FILE *oldf;
int SOLVE;

  name = varTable[ f ]->name;
  narg = varTable[ f ]->maxi;
    
  Va_start( args, f );
  for( i = 0; i < narg; i++ ) 
    vars[ i ] = va_arg( args, int );
  va_end( args );
    
  CommentFncBegin( f, vars );
  F95_FunctionStart( f, vars );
  NewLines(1);
  bprintf("      USE SPRSDATA\n");
  bprintf("      IMPLICIT NONE\n");/* mz_rs_20030728 */

  FlushBuf();

  for( i = 0; i < narg; i++ ) 
    F95_Declare( vars[ i ] );

  bprintf("\n");
  FlushBuf();

  MapFunctionComment( f, vars );
}

void F95_FunctionEnd( int f )		/* F95 line */
{
  bprintf("      END SUBROUTINE %s\n\n", varTable[ f ]->name );

  FlushBuf();

  CommentFunctionEnd( f );
}

void F95_Inline( char *fmt, ... )	/* no change */
{
Va_list args;
char buf[ 1000 ];

  if( useLang != F95_LANG ) return;
  
  Va_start( args, fmt );
  vsprintf( buf, fmt, args );
  va_end( args );
  bprintf( "%s\n", buf );
  
  FlushBuf();
}

/* mz_rs_20030726: F95_Second_WriteVecData deleted */

char * F95_Second_Decl( int v, int n, void * values)
{
int i, j;
int nlines, min, max;
int split;
static char buf[120];
VARIABLE *var;
int * ival;
double * dval;
char ** cval;
char *baseType;
char maxi[20];
char maxj[20];
int maxCols = MAX_COLS;
 
 int i_from, i_to;
 int isplit;
 int splitsize;
 int maxi_mod;
 int maxi_div;

  var = varTable[ v ];
  ival = (int*) values;
  dval = (double *) values;
  cval = (char **) values;

  nlines = 1;
  min = max = 1;
  split = 0;
  var -> maxi = max( n, 1 );

  baseType = F95_types[ var->baseType ];
  
  *buf = 0;

  switch( var->type ) {				/* changed here */
    case ELM:   
	    bprintf( "      %s :: %s = ", baseType, var->name );
		switch ( var->baseType ) {
		  case INT: bprintf( "%d", *ival ); break;
		  case DOUBLE: bprintf( "%f", *dval); break;
		  case REAL: bprintf( "%lg", *dval ); break;
		  case STRING: bprintf( "'%3s'", *cval ); break;
		}
		break;
    case VELM:
      /* mz_rs_20030726: many changes in case VELM section */
      /* mz_rs_20030726: var->maxi and n seem to be identical (but when
         var->maxi=1, then n=0 !!!) */
      splitsize = 360;
      maxi_mod = var->maxi % splitsize;
      maxi_div = var->maxi / splitsize;
      /* mz_rs_20040430+ */
      /* correction if var->maxi is a multiple of splitsize */
      if ( (maxi_div>0) && (maxi_mod==0) ) {
        maxi_mod = splitsize;
        maxi_div--;
      }
      /* mz_rs_20040430- */
      for ( isplit=0; isplit <= maxi_div; isplit++ ) {
        if( var->maxi > 0 ) sprintf( maxi, "%d", var->maxi );
        else sprintf( maxi, "%s", varTable[ -var->maxi ]->name );  
        if( (var->maxi == 0) || 
            ((var->maxi < 0) && (varTable[ -var->maxi ]->maxi == 0)) )
          strcat( maxi, "+1");
        bprintf( "      %s, " , baseType);
        if( n>0 ) bprintf( "PARAMETER, " ); /* if values are assigned now */
        if ( maxi_div==0 ) /* define array in one piece */
          bprintf( "DIMENSION(%s) :: %s", 
                   maxi, var->name) ;
        else /* define partial arrays */
          if ( isplit==maxi_div ) /* last part has size maxi_mod */
            bprintf( "DIMENSION(%d) :: %s_%d", 
                     maxi_mod, var->name, isplit) ;
          else /* all other parts have size splitsize */
            bprintf( "DIMENSION(%d) :: %s_%d", 
                     splitsize, var->name, isplit) ;
        if( n<=0 ) break;

        /* now list values */
        bprintf( " &\n       = (/" );
        /* if the array is defined in one piece, then the for loop will
           go from 0 to n. Otherwise, there will be partial arrays from
           i_from to i_to which are of size splitsize except for the
           last one which is usually smaller and contains the rest */
        i_from = isplit * splitsize;
        i_to   = MIN(i_from+splitsize,n);
        for ( i=i_from; i < i_to; i++ ) {
          switch( var -> baseType ) {
          case INT:
            bprintf( "%3d", ival[i] ); maxCols =12; break;
          case DOUBLE:
            bprintf( "%5f", dval[i] ); maxCols = 8; break;
          case REAL:
            bprintf( "%5lg", dval[i] ); maxCols = 8; break;
          case STRING:
            bprintf( "'%3s'", cval[i] ); maxCols = 5; break;
          }
          if( i < i_to-1 ) {
            bprintf( "," );
            if( (i+1) % maxCols == 0 ) {
              bprintf( " &\n     " );
              nlines++;
            }
          }
          max++;
        }
        bprintf( " /)\n" );
      }

      /* combine the partial arrays */
      if ( maxi_div != 0 ) {
        bprintf( "  %s, PARAMETER, DIMENSION(%s) :: %s = (/&\n    ", 
                 baseType, maxi, var->name) ;
        for ( isplit=0; isplit <= maxi_div; isplit++ ) {
          bprintf( "%s_%d", var->name, isplit) ;
          if( isplit < maxi_div ) { /* more parts will follow */
            bprintf( ", " );
            /* line break after 5 variables */
            if( (isplit+1) % 5 == 0 ) bprintf( "&\n    " );
          } else { /* after last part */
            bprintf( " /)\n" );
          }
        }
      }

      break;
				
    case MELM:  if( var->maxi > 0 ) sprintf( maxi, "%d", var->maxi );
                  else sprintf( maxi, "%s", varTable[ -var->maxi ]->name );  
                if( (var->maxi == 0) || 
                    ((var->maxi < 0) && (varTable[ -var->maxi ]->maxi == 0)) )
                  strcat( maxi, "+1");
                if( var->maxj > 0 ) sprintf( maxj, "%d", var->maxj );
                  else sprintf( maxj, "%s", varTable[ -var->maxj ]->name );  
                if( (var->maxj == 0) || 
                    ((var->maxj < 0 ) && (varTable[ -var->maxj ]->maxi == 0)) )
                  strcat( maxj, "+1");
                sprintf( buf, "%s, DIMENSION(%s,%s) :: %s\n",	/* changed here */		
                         baseType, maxi, maxj,var->name );  
		break;
    default:
                printf( "Can not declare type %d", var->type );
                break;
  }
  return buf;
}

void F95_Second_Declare( int v, int n, void * values)
{
  if ( varTable[ v ] -> comment ) {
    F95_WriteComment( "%s - %s",
		    varTable[ v ] -> name, varTable[ v ]->comment );
  }
  bprintf("      %s\n", F95_Second_Decl( v, n, values ) );

  FlushBuf();
}

void F95_Second_ExternDeclare( int v, int n, void * values )
{
  F95_Second_Declare( v, n, values );
}


void Use_F95()				/* changed function names to F95 */
{ 
  WriteElm 	    = F95_WriteElm;
  WriteSymbol 	    = F95_WriteSymbol;  
  WriteAssign 	    = F95_WriteAssign;
  WriteComment 	    = F95_WriteComment;
  DeclareConstant   = F95_DeclareConstant;
  Declare           = F95_Declare;
  ExternDeclare     = F95_ExternDeclare;

  SecondExtern	    = F95_Second_ExternDeclare;
  
  GlobalDeclare     = F95_GlobalDeclare;
  InitDeclare       = F95_InitDeclare;

  FunctionStart     = F95_FunctionStart;
  FunctionPrototipe = F95_FunctionPrototipe;
  FunctionBegin     = F95_FunctionBegin;
  FunctionBeginSprs = F95_FunctionBeginSprs;
  FunctionEnd       = F95_FunctionEnd;

  /* OpenFile( &hFile,   rootFileName, ".h" ); mz_rs_20020426: deleted */
  OpenFile( &srcFile, rootFileName, ".f90" );
  /* OpenFile( &sparseFile, rootFileName, "_s.h" ); mz_rs_20020426: deleted */
  OpenFile( &mapFile, rootFileName, ".map" );
  OpenFile( &sparseModuleFile, rootFileName, "_module_s.f90" );
  OpenFile( &ModuleHeadFile, rootFileName, "_module.f90" );
} 
